"""
A utility function to be called when optimizing reward configurations
"""
import logging
import os
from collections import defaultdict
from pprint import pformat

import numpy as np
import optuna
import yaml
from optuna.integration.skopt import SkoptSampler
from optuna.pruners import SuccessiveHalvingPruner, MedianPruner
from optuna.samplers import RandomSampler, TPESampler
from stable_baselines.common.base_class import _UnvecWrapper
from stable_baselines.common.vec_env import VecEnv
from stable_baselines.her import HERGoalEnvWrapper

from commonroad_rl.plot_learning_curves import ts2collision, ts2reward, ts2goal, ts2off_road, ts2max_time
from commonroad_rl.utils_run.callbacks import RewardConfigsTrialEvalCallback
from commonroad_rl.utils_run.plot_util import default_xy_fn
from commonroad_rl.utils_run.plot_util import load_results as load_results_baselines, default_split_fn

__author__ = "Brian Liao, Johannes Kaiser"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Xiao Wang"
__email__ = "xiao.wang@tum.de"
__status__ = "Released"

# Methods that retrieve the individual parameters of monitor.csv
PLOT_DICT = {
    # "Mean Reward": ts2ep_rew_mean,
    "Total Reward": ts2reward, # ts2xy,
    # "Total Robustness reward": ts2monitor_reward,
    # "Total Sparse reward": ts2gym_reward,
    # "Step Robustness reward": ts2monitor_reward_step,
    # "Step Sparse reward": ts2gym_reward_step,
    "Goal-Reaching Rate": ts2goal,
    "Collision Rate": ts2collision,
    "Off-Road Rate": ts2off_road,
    "Max Time Reached": ts2max_time,
}

""" 
Dict used for the guided optimization
Structure: 
Failure rate
    | reward type
    |   | [list of considered rewards for the failure and reward type]
    ...
"""
GUIDED = {
    "Off-Road Rate":{
        "reward_configs_dense":[],
        "reward_configs_hybrid":["reward_off_road"],
        "reward_configs_sparse":[]
    },
    "Max Time Reached":{
        "reward_configs_dense":[],
        "reward_configs_hybrid":["reward_time_out"],
        "reward_configs_sparse":[]
    },
    "Collision Rate":{
        "reward_configs_dense":[],
        "reward_configs_hybrid":["reward_collision"],
        "reward_configs_sparse":[]
    },
    "Goal-Reaching Rate":{
        "reward_configs_dense":[],
        "reward_configs_hybrid":["reward_goal_reached"],
        "reward_configs_sparse":[]
    }
}

"""
Dict of all considered reward types for the guided optimization with 
respective [direction of improvement, rate aimed for]
ex. reward_off_road improves the goal reching rate when it is decreased
    the goal off_road_rate is 0 when considering reward_off_road 
"""

# first entry of array is direction of improvement of related rewards with scaling
# second entry of array is goal value
SIGN_OF_IMPROVEMENT = {
    "reward_off_road" : [-10, 0],
    "reward_time_out" : [-10, 0],
    "reward_collision" : [-10, 0],
    "reward_goal_reached" : [10, 1],
}


# Value the rewards are normalized to
SUM_OF_REWARD_VALUES = 300


def parse_sampler(sampler_method: str, n_startup_trials: int, seed: int):
    """
    This function loads an enviromenet based on the env_configs

    :param sampler_method: string of [random, tpe, skopt]
    :param n_startup_trials: int 
    :param seed: int seeding the sampling for reproducing results
    """
    # n_warmup_steps: Disable pruner until the trial reaches the given number of step.
    if sampler_method == "random":
        return RandomSampler(seed=seed)
    elif sampler_method == "tpe":
        return TPESampler(n_startup_trials=n_startup_trials, seed=seed)
    elif sampler_method == "skopt":
        # cf https://scikit-optimize.github.io/#skopt.Optimizer
        # GP: gaussian process
        # Gradient boosted regression: GBRT
        return SkoptSampler(skopt_kwargs={"base_estimator": "GP", "acq_func": "gp_hedge"})
    else:
        raise ValueError(f"[reward_configs_opt.py] Unknown sampler: {sampler_method}")


def parse_pruner(pruner_method: str, n_startup_trials: int, n_evaluations: int, n_trials: int):
    """
    This function loads an enviromenet based on the env_configs

    :param pruner_method: string of [halving, median, none]
    :param n_startup_trials: int 
    :param n_evaluations: int number of evaluations of the model after each eval_freq
    :param  n_trials: int number of trials to be executed
    """
    if pruner_method == "halving":
        return SuccessiveHalvingPruner(min_resource=1, reduction_factor=4, min_early_stopping_rate=0)
    elif pruner_method == "median":
        return MedianPruner(n_startup_trials=n_startup_trials, n_warmup_steps=n_evaluations // 3)
    elif pruner_method == "none":
        # Do not prune
        return MedianPruner(n_startup_trials=n_trials, n_warmup_steps=n_evaluations)
    else:
        raise ValueError(f"[reward_configs_opt.py] Unknown pruner: {pruner_method}")


def average(lst):
    return sum(lst)/len(lst)


def smooth_results(allresults, xy_fn=default_xy_fn, split_fn=default_split_fn, group_fn=default_split_fn):
    """
    Snoothes the result over the last few evaluations

    :param allresults: (float)
    :param xy_fn: function Result -> x,y           - function that converts results objects into tuple of x and y values.
                                              By default, x is cumsum of episode lengths, and y is episode rewards

    :param split_fn: function Result -> hashable   - function that converts results objects into keys to split curves into
                                              sub-panels by.
                                              That is, the results r for which split_fn(r) is different will be put
                                              on different sub-panels.
                                              By default, the portion of r.dirname between last / and -<digits> is
                                              returned. The sub-panels are
                                              stacked vertically in the figure.

    :param group_fn: function Result -> hashable   - function that converts results objects into keys to group curves by.
                                              That is, the results r for which group_fn(r) is the same will be put
                                              into the same group.
                                              Curves in the same group have the same color (if average_group is
                                              False), or averaged over
                                              (if average_group is True). The default value is the same as default
                                              value for split_fn
    :return: (list) ordered list of averaged results 
    """
    
    result_list = []

    if split_fn is None: split_fn = lambda _: ''
    if group_fn is None: group_fn = lambda _: ''
    sk2r = defaultdict(list)  # splitkey2results
    for result in allresults:
        splitkey = split_fn(result)
        sk2r[splitkey].append(result)
    assert len(sk2r) > 0

    for (isplit, sk) in enumerate(sorted(sk2r.keys())):
        g2c = defaultdict(int)
        sresults = sk2r[sk]
        for result in sresults:
            group = group_fn(result)
            g2c[group] += 1
            x, y = xy_fn(result)
            if x is None: x = np.arange(len(y))
            x, y = map(np.asarray, (x, y))

            if len(x) >= 10:
                avg_y = average(y[-int(len(y)/3):])
            elif len(x) != 0:
                avg_y = y[-1]
            else:
                avg_y = 0

            result_list.append(avg_y)
    return average(result_list)


def group_fn(results):
    return os.path.basename(results.dirname)


def optimize_reward_configs(
        algo,
        env_id,
        model_fn,
        env_fn,
        sampling_setting,
        n_timesteps=5000,
        eval_freq=500,
        n_eval_episodes=5,
        n_trials=10,
        hyperparams=None,
        configs=None,
        n_jobs=1,
        sampler_method="random",
        pruner_method="halving",
        seed=13,
        verbose=1,
        log_path=None,
        best_model_save_path=None,
        guided=False,
):
    """

    :param algo: (str)
    :param env_id: (str)
    :param model_fn: (func) function that is used to instantiate the model
    :param env_fn: (func) function that is used to instantiate the env
    :param n_trials: (int) maximum number of trials for finding the best hyperparams
    :param n_timesteps: (int) maximum number of timesteps per trial
    :param hyperparams: (dict) model hyperparameters
    :param configs: (dict) environment configurations to be optimized
    :param n_jobs: (int) number of parallel jobs
    :param sampler_method: (str)
    :param pruner_method: (str)
    :param seed: (int)
    :param sampling_setting: (dict) sampling intervals for correspinding items
    :param verbose: (int)
    :param log_path: (str) folder for saving evaluation results during optimization
    :param best_model_save_path (str) folder for saving the best model
    :param guided: (bool) if True, the optimization is performed in a guided manner
    :return: (dict) detailed result of the optimization
    """
    # TODO: remove; duplicate in RewardConfigsTrialEvalCallback
    # LOGGER.setLevel(verbose)
    #
    # if not len(LOGGER.handlers):
    #     formatter = logging.Formatter("[%(levelname)s] %(name)s - %(message)s")
    #     stream_handler = logging.StreamHandler()
    #     stream_handler.setLevel(verbose)
    #     stream_handler.setFormatter(formatter)
    #     LOGGER.addHandler(stream_handler)
    #
    #     if log_path is not None:
    #         file_handler = logging.FileHandler(filename=os.path.join(log_path, "console_copy.txt"))
    #         file_handler.setLevel(verbose)
    #         file_handler.setFormatter(formatter)
    #         LOGGER.addHandler(file_handler)

    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler(os.path.join(log_path, "console_copy2.txt"))
    file_handler.setLevel(logging.INFO)
    LOGGER.addHandler(file_handler)

    LOGGER.info(f"Optimizing environment configurations for {env_id}")

    # Enviornment configurations including observation space settings and reward weights
    if configs is None:
        configs = {}

    n_startup_trials = 10
    n_evaluations = n_timesteps // eval_freq

    LOGGER.info(f"Optimizing with {n_trials} trials using {n_jobs} parallel jobs, "
                f"each with {n_timesteps} maximal time steps and {n_evaluations} evaluations")

    sampler = parse_sampler(sampler_method, n_startup_trials, seed)
    pruner = parse_pruner(pruner_method, n_startup_trials, n_evaluations, n_trials)
    LOGGER.debug(f"Sampler: {sampler_method} - Pruner: {pruner_method}")

    # Create study object on environment configurations from Optuna
    reward_configs_study = optuna.create_study(sampler=sampler, pruner=pruner)

    trial_dict = {}
    based_on_dict = {}

    def objective_reward_configs(trial):
        """
        Optimization objective for environment configurations
        """
        LOGGER.info(f"Trial dict: {trial_dict}")
        LOGGER.info(f"Based_on dict: {based_on_dict}")
        trial.model_class = None

        if algo == "her":
            trial.model_class = hyperparams["model_class"]

        # Hack to use DDPG/TD3 noise sampler
        if algo in ["ddpg", "td3"] or trial.model_class in ["ddpg", "td3"]:
            trial.n_actions = env_fn(n_envs=1).action_space.shape[0]

        # Get keyword arguments for model hyperparameters and environment configurations
        kwargs_hyperparams = hyperparams.copy()
        kwargs_configs = configs.copy()

        # Sample environment configurations and keep model hyperparameters
        sampled_reward_configs = sample_commonroad_reward_configs(trial, sampling_setting, trial_dict, based_on_dict, log_path, LOGGER, guided)
        kwargs_configs.update(sampled_reward_configs)

        # Save data for later inspection
        tmp_path = os.path.join(log_path, "trial_" + str(trial.number))
        os.makedirs(tmp_path, exist_ok=True)
        with open(os.path.join(tmp_path, "sampled_reward_configurations.yml"), "w") as f:
            yaml.dump(kwargs_configs, f)
            LOGGER.info("Saving sampled reward configurations into " + tmp_path)
        LOGGER.debug("[reward_configs_opt.py] Sampled reward configurations:")
        LOGGER.debug(pformat(kwargs_configs))

        # Create model and environments for optimization TODO: save model in tmp_path
        model = model_fn(kwargs_hyperparams, kwargs_configs)
        # TODO: why n_envs=1 for eval_env --> EvalCallback only allows one eval_env --> Use MultiEvalCallback instead??
        eval_env = env_fn(n_envs=1, eval_env=True, **kwargs_configs)

        # Account for parallel envs
        model_env = model.get_env()
        if isinstance(model_env, VecEnv):
            eval_freq_ = max(eval_freq // model_env.num_envs, 1)

        LOGGER.info(f"Evaluating with {n_eval_episodes} episodes after every {eval_freq_} time steps")

        reward_configs_eval_callback = RewardConfigsTrialEvalCallback(
            eval_env,
            trial,
            n_eval_episodes=n_eval_episodes,
            eval_freq=eval_freq_,
            log_path=log_path,
            best_model_save_path=best_model_save_path,
            deterministic=True,
            verbose=verbose,
        )

        if algo == "her":
            # Wrap the env if need to flatten the dict obs
            if isinstance(eval_env, VecEnv):
                eval_env = _UnvecWrapper(eval_env)
            eval_env = HERGoalEnvWrapper(eval_env)

        try:
            # TODO: what is model.runner?
            model.learn(n_timesteps, callback=reward_configs_eval_callback)
            # Free memory
            model.save(os.path.join(tmp_path, "trial_" + str(trial.number)))
            model.env.close()
            eval_env.close()
            
        except AssertionError:
            # Sometimes, random hyperparams can generate NaN
            # Free memory
            model.env.close()
            eval_env.close()
            raise optuna.exceptions.TrialPruned()

        past_results = load_results_baselines(os.path.dirname(log_path))
        own_trial_dict = {}
        for i, (k, xy_fn) in enumerate(PLOT_DICT.items()):
            output = smooth_results(past_results, group_fn=group_fn, xy_fn=xy_fn)
            own_trial_dict[k] = output
        trial_dict["trial_" + str(trial.number)] = own_trial_dict

        is_pruned = reward_configs_eval_callback.is_pruned
        cost = reward_configs_eval_callback.cost
        del model.env, eval_env
        del model
        if is_pruned:
            raise optuna.exceptions.TrialPruned()
        LOGGER.info(f"Cost of {trial.number} resulted in {cost}")
        return cost

    try:
        LOGGER.info(f"Trying to optimize environment configurations with {n_trials} trials and {n_jobs} jobs")
        reward_configs_study.optimize(objective_reward_configs, n_trials=n_trials, n_jobs=n_jobs)
        LOGGER.info(f"final trial dict: {trial_dict}")
    except KeyboardInterrupt:
        pass

    LOGGER.info(f"Number of finished trials: {len(reward_configs_study.trials)}")
    LOGGER.info(f"Best value: {reward_configs_study.best_trial.value}")
    LOGGER.info("Best reward configurations: ")
    LOGGER.info(pformat(reward_configs_study.best_trial.params))

    return reward_configs_study.best_trial.params


def sample_commonroad_reward_configs(trial, sampling_setting, trial_dict, based_on_dict, log_path, logger, guided):
    # Read in reward items and intervals, and generate random values
    reward_configs = {}
    for key, values in sampling_setting.items():
        if len(values) == 1:
            method, interval = next(iter(values.items()))
            if method == "categorical":
                reward_configs[key] = trial.suggest_categorical(key, interval)
            elif method == "uniform":
                reward_configs[key] = trial.suggest_uniform(key, interval[0], interval[1])
            elif method == "loguniform":
                reward_configs[key] = trial.suggest_loguniform(key, interval[0], interval[1])
            else:
                print(f"[reward_configs_opt.py] Sampling method {method} not supported")
        elif len(values) > 1:
            # TODO: commonroad-v1
            reward_configs[key] = dict()
            for subkey, subvalue in values.items():
                method, interval = next(iter(subvalue.items()))
                if method == "categorical":
                    reward_configs[key][subkey] = trial.suggest_categorical(subkey, interval)
                elif method == "uniform":
                    reward_configs[key][subkey] = trial.suggest_uniform(subkey, interval[0], interval[1])
                elif method == "loguniform":
                    reward_configs[key][subkey] = trial.suggest_loguniform(subkey, interval[0], interval[1])
                else:
                    print(f"[reward_configs_opt.py] Sampling method {method} not supported")
    
    if guided: # guided optimization
        logger.info("guided trial")
        compare_trial = trial.number - 1
        updated = False
        while not updated:
            if compare_trial == -1:
                # First trial
                # Do as normal, as first trial or first trial incomplete
                # Trial dict based on is none
                based_on_dict["trial_" + str(trial.number)] = None
                # updated allows to stop looking for a comparisson trial
                updated = True
            elif "trial_" + str(compare_trial) in trial_dict:  # one trial before current is found

                previous_trial = "trial_" + str(compare_trial)
                # based on dict is updated
                based_on_dict["trial_" + str(trial.number)] = previous_trial
                previous_rates = trial_dict[previous_trial]
                previous_rewards_path = os.path.join(os.path.join(log_path, previous_trial), "sampled_reward_configurations.yml")
                previous_rewards = yaml.load(open(previous_rewards_path))

                if based_on_dict[previous_trial] != None:   # second trial before current is found
                    second_to_prev_trial = based_on_dict[previous_trial]
                    second_to_prev_rates = trial_dict[second_to_prev_trial]
                    second_to_prev_rewards_path = os.path.join(os.path.join(log_path, second_to_prev_trial), "sampled_reward_configurations.yml")
                    second_to_prev_rewards = yaml.load(open(second_to_prev_rewards_path))
                    for failure, value in GUIDED.items():   # for each type of failure: off_road etc.
                        for key, subkeys in value.items():  # for each reward type
                            for subkey in subkeys:          # for each reward
                                # the new reward is the sum of the past reward, a change based on the failure rate, a change based on the change of failure rate with respect to the last change in reward
                                # SIGN_OF_IMPROVEMENT[subkey][1] is used to guide the results in different directions: goal reaching rate --> 1; off road rate --> 0
                                # SIGN_OF_IMPROVEMENT[subkey][0] to specify the direction of improvement goal_reaching rate getting bigger forces the model to reach the goal, off_road reward getting smaller forces the model to stay on the road
                                change_in_reward_value = previous_rewards[key][subkey] - second_to_prev_rewards[key][subkey] + 0.000001 #change reward
                                change_in_failure = abs(SIGN_OF_IMPROVEMENT[subkey][1] - second_to_prev_rates[failure]) - abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure]) #positive if improvement
                                # If we get close to the optimum drop failure rate based change and rely on derivative based change
                                if abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure]) > 0.1:
                                    scaling =  abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure])
                                else:
                                    scaling = 0
                                # allow for bigger jumps when far away from optimum, go to local minimum afterwards
                                reward_configs[key][subkey] = float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * (- change_in_failure / np.sign(change_in_reward_value) + scaling * 3))
                                logger.info(f"reward {subkey} set to {float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * (- change_in_failure / np.sign(change_in_reward_value) + scaling * 3))}")
                                logger.debug(f" failure {failure}")
                                logger.debug(f"rates {previous_rates[failure]}")
                                logger.debug(f"adapted rates {abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure])}")
                                logger.debug(f"new rewards {float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * (- change_in_failure / np.sign(change_in_reward_value) + scaling * 3))}")
                                logger.debug(f"change {float(int(SIGN_OF_IMPROVEMENT[subkey][0]) * (- change_in_failure / np.sign(change_in_reward_value) + scaling * 3))}")
                else:   # Only on comparisson trial is found (Only one finished yet)
                    for failure, value in GUIDED.items():
                        for key, subkeys in value.items():
                            for subkey in subkeys:
                                # No derivative if only one trial can be found
                                reward_configs[key][subkey] = float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure]) * 3)
                                logger.info(f"reward {subkey} set to {float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure]) * 3)}")
                                logger.debug(f" failure {failure}")
                                logger.debug(f"rates {previous_rates[failure]}")
                                logger.debug(f"new rewards {float(previous_rewards[key][subkey] + int(SIGN_OF_IMPROVEMENT[subkey][0]) * abs(SIGN_OF_IMPROVEMENT[subkey][1] - previous_rates[failure]) * 3)}")
                #  updated allows to stop looking for a comparisson trial
                updated = True
            else:
                compare_trial -= 1  # Try to find trial with one number less, allows for multithreading, when process 5 is started, we can still compare to results of trial 1
        logger.info(f"reward configs before norm {reward_configs}")
        # Norming
        total_reward_of_optimized = 0
        for failure, value in GUIDED.items():   # for each type of failure: off_road etc.
                for key, subkeys in value.items():  # for each reward type
                    for subkey in subkeys:
                        total_reward_of_optimized = total_reward_of_optimized + abs(reward_configs[key][subkey])
        for failure, value in GUIDED.items():   # for each type of failure: off_road etc.
                for key, subkeys in value.items():  # for each reward type
                    for subkey in subkeys:
                        reward_configs[key][subkey] = (reward_configs[key][subkey]/total_reward_of_optimized)*SUM_OF_REWARD_VALUES
    logger.info(f"reward configs {reward_configs}")
    return reward_configs


REWARD_CONFIGS_SAMPLER = {
    "commonroad-v0": sample_commonroad_reward_configs,
    "commonroad-v1": sample_commonroad_reward_configs
}



import os
import glob
import random
import ntpath
import argparse

from shutil import copyfile


def get_args():
    parser = argparse.ArgumentParser(description="Divide scenarios into test and training scenarios",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--input_dir", "-i", type=str,
                        default="/home/wangx/data/highD-dataset-v1.0/cr_scenarios/pickle/problem")
    parser.add_argument("--output_dir", "-o", type=str,
                        default="/home/wangx/data/highD-dataset-v1.0/cr_scenarios/pickle")

    return parser.parse_args()


def main():
    args = get_args()

    train_path = os.path.join(args.output_dir, "problem_train")
    test_path = os.path.join(args.output_dir, "problem_test")

    os.makedirs(train_path, exist_ok=True)
    os.makedirs(test_path, exist_ok=True)

    fns = sorted(glob.glob(os.path.join(args.input_dir, "*.pickle")))
    random.shuffle(fns)

    num_train = int(len(fns) * 0.7)

    print("Copying training data ...")
    for i, fn in enumerate(fns[:num_train]):
        print(f"{i + 1}/{num_train}", end="\r")
        copyfile(fn, os.path.join(train_path, ntpath.basename(fn)))

    print("Copying test data...")
    for i, fn in enumerate(fns[num_train:]):
        print(f"{i + 1}/{len(fns) - num_train}", end="\r")
        copyfile(fn, os.path.join(test_path, ntpath.basename(fn)))


if __name__ == "__main__":
    main()



"""
Script to filter out scenarios where no route can be found
"""

__author__ = "Armin Ettenhofer"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Armin Ettenhofer"
__email__ = "armin.ettenhofer@tum.de"
__status__ = "Released"

import argparse
import logging
import os
import pickle

from commonroad.planning.planning_problem import PlanningProblem
from commonroad_route_planner.route_planner import RoutePlanner

import commonroad_rl.gym_commonroad.constants as constants
from commonroad_rl.gym_commonroad.utils.scenario_io import restore_scenario

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
LOGGER.addHandler(handler)


def get_args():
    """Scan arguments"""
    parser = argparse.ArgumentParser(
        description="Filters out non-routable scenarios",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--pickles", "-i", help="Path to problems", type=str,
                        default=constants.PATH_PARAMS["pickles"])
    parser.add_argument("--output", "-o", help="Path to output dir", type=str,
                        default=os.path.join(constants.PATH_PARAMS["pickles"], "unroutable-problems"))
    return parser.parse_args()


def main(pickles: str, output: str) -> None:
    """
    Filter out non-routable scenarios
    """
    pickles = os.path.expanduser(pickles)
    output = os.path.expanduser(output)
    os.makedirs(name=output, exist_ok=True)
    problem = os.path.join(pickles, "problem")
    meta_scenario_path = os.path.join(pickles, "meta_scenario")

    with open(os.path.join(meta_scenario_path, "problem_meta_scenario_dict.pickle"), "rb") as f:
        problem_meta_scenario_dict = pickle.load(f)

    for file in os.listdir(problem):
        if not os.path.isfile(os.path.join(problem, file)) or not file.endswith(".pickle"):
            continue

        with open(os.path.join(problem, file), "rb") as f:
            problem_dict: dict = pickle.load(f)

        planning_problem: PlanningProblem = list(problem_dict["planning_problem_set"].planning_problem_dict.values())[0]
        benchmark_id = os.path.splitext(file)[0]
        meta_scenario = problem_meta_scenario_dict[benchmark_id]
        scenario = restore_scenario(meta_scenario, problem_dict["obstacle"])

        rp = RoutePlanner(scenario, planning_problem)
        routes = rp.plan_routes()

        if routes.num_route_candidates == 0:
            os.rename(os.path.join(problem, file), os.path.join(output, file))


if __name__ == "__main__":
    args = get_args()
    main(
        pickles=args.pickles,
        output=args.output
    )



import logging
import os
import warnings
from typing import Union, Optional, Callable, Tuple, List

import gym
import numpy as np
from stable_baselines.common.callbacks import BaseCallback, EvalCallback, EventCallback
from stable_baselines.common.vec_env import (
    VecEnv,
    DummyVecEnv,
    sync_envs_normalization,
)

LOGGER = logging.getLogger(__name__)
# from commonroad_rl.train_model import construct_logger
# dirty fix, TODO: investivate why can not import from train_model.py
def construct_logger(logging_mode: int, save_path: str, logger):
    logger.setLevel(logging_mode)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging_mode)
    file_handler = logging.FileHandler(os.path.join(save_path, "console_copy.txt"))
    file_handler.setLevel(logging_mode)

    formatter = logging.Formatter("[%(levelname)s] %(name)s - %(message)s")
    stream_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)
    logger.debug(f"Logging to console and {save_path + '/console_copy.txt'}")


def construct_path_with_trial_number(path: str, trial_numer):
    if path is not None:
        if trial_numer == "best_model":
            return os.path.join(path, trial_numer)
        else:
            return os.path.join(path, "trial_" + str(trial_numer))
    else:
        return path


class HyperparamsTrialEvalCallback(EvalCallback):
    """
    Callback used for evaluating and reporting a trial during model hyperparameter optimization.
    """

    def __init__(self, eval_env, trial, n_eval_episodes=5, eval_freq=10000, log_path=None,
                 best_model_save_path=None, deterministic=True, verbose=1):

        best_model_save_path = construct_path_with_trial_number(best_model_save_path, trial.number)
        log_path = construct_path_with_trial_number(log_path, trial.number)

        super(HyperparamsTrialEvalCallback, self).__init__(
            eval_env=eval_env,
            n_eval_episodes=n_eval_episodes,
            eval_freq=eval_freq,
            log_path=log_path,
            best_model_save_path=best_model_save_path,
            deterministic=deterministic,
            verbose=verbose,
        )
        self.trial = trial
        self.eval_idx = 0
        self.is_pruned = False
        self.cost = 0.0

    def _on_step(self):
        if self.eval_freq > 0 and self.n_calls % self.eval_freq == 0:
            super(HyperparamsTrialEvalCallback, self)._on_step()
            self.eval_idx += 1
            # report best or report last?
            # report num_timesteps or elasped time?
            self.cost = -1 * self.best_mean_reward
            self.trial.report(self.cost, self.eval_idx)
            # Prune trial if need
            if self.trial.should_prune():
                self.is_pruned = True
                return False
        return True


class ObservationConfigsTrialEvalCallback(EvalCallback):
    """
    Callback used for evaluating and reporting a trial during observation configuration optimization.
    """

    def __init__(
            self,
            eval_env,
            trial,
            n_eval_episodes=5,
            eval_freq=10000,
            log_path=None,
            best_model_save_path=None,
            deterministic=True,
            verbose=1,
    ):
        best_model_save_path = construct_path_with_trial_number(best_model_save_path, trial.number)
        log_path = construct_path_with_trial_number(log_path, trial.number)

        super(ObservationConfigsTrialEvalCallback, self).__init__(
            eval_env=eval_env,
            n_eval_episodes=n_eval_episodes,
            eval_freq=eval_freq,
            log_path=log_path,
            best_model_save_path=best_model_save_path,
            deterministic=deterministic,
            verbose=verbose,
        )
        self.trial = trial
        self.eval_idx = 0
        self.is_pruned = False
        self.cost = 0.0

    def _on_step(self):
        if self.eval_freq > 0 and self.n_calls % self.eval_freq == 0:
            super(ObservationConfigsTrialEvalCallback, self)._on_step()
            self.eval_idx += 1
            self.cost = -1 * self.best_mean_reward
            self.trial.report(self.cost, self.eval_idx)
            # Prune trial if need
            if self.trial.should_prune():
                self.is_pruned = True
                return False
        return True


class RewardConfigsTrialEvalCallback(EvalCallback):
    """
    Callback used for evaluating and reporting a trial during reward configaration optimization.
    """

    def __init__(
            self,
            eval_env,
            trial,
            n_eval_episodes=5,
            eval_freq=1000,
            log_path=None,
            best_model_save_path=None,
            deterministic=True,
            verbose=1,
    ):
        # Set save path and logger
        map_verbose_to_logging = {2: logging.DEBUG, 1: logging.INFO}
        self.log_path = construct_path_with_trial_number(log_path, trial.number)
        # Save best model into `($best_model_save_path)/trial_($trial_number)/best_model.zip`
        self.best_model_save_path = construct_path_with_trial_number(best_model_save_path, trial.number)
        construct_logger(map_verbose_to_logging.get(verbose, logging.ERROR), self.log_path, LOGGER)

        super(RewardConfigsTrialEvalCallback, self).__init__(eval_env=eval_env,
                                                             n_eval_episodes=n_eval_episodes,
                                                             eval_freq=eval_freq,
                                                             deterministic=deterministic,
                                                             verbose=verbose)
        self.trial = trial
        self.eval_idx = 0
        self.is_pruned = False
        self.lowest_mean_cost = np.inf
        self.last_mean_cost = np.inf
        self.cost = 0.0

        # Log evaluation information into `($log_path)/trial_($trial_number)/evaluations.npz`
        self.evaluation_timesteps = []
        self.evaluation_costs = []
        self.evaluation_lengths = []

    def _on_step(self):
        if self.eval_freq > 0 and self.n_calls % self.eval_freq == 0:

            def evaluate_policy_configs(model, env, n_eval_episode=10, render=False, deterministic=True, callback=None):
                """
                Runs policy for `n_eval_episodes` episodes and returns cost for optimization.
                This is made to work only with one env.

                :param model: (BaseRLModel) The RL agent you want to evaluate.
                :param env: (gym.Env or VecEnv) The gym environment. In the case of a `VecEnv`, this must contain only one environment.
                :param n_eval_episode: (int) Number of episode to evaluate the agent
                :param deterministic: (bool) Whether to use deterministic or stochastic actions
                :param render: (bool) Whether to render the environment or not
                :param callback: (callable) callback function to do additional checks, called after each step.
                :return: ([float], [int]) list of episode costs and lengths
                """
                if isinstance(env, VecEnv):
                    assert env.num_envs == 1, "You must pass only one environment when using this function"

                episode_costs = []
                episode_lengths = []
                for _ in range(n_eval_episode):
                    obs = env.reset()
                    done, info, state = False, None, None

                    # Record required information
                    # Since vectorized environments get reset automatically after each episode,
                    # we have to keep a copy of the relevant states here.
                    # See https://stable-baselines.readthedocs.io/en/master/guide/vec_envs.html for more details.
                    episode_length, episode_cost = 0, 0.
                    episode_is_time_out = []
                    episode_is_collision = []
                    episode_is_off_road = []
                    episode_is_goal_reached = []
                    episode_is_friction_violation = []
                    while not done:
                        action, state = model.predict(obs, state=state, deterministic=deterministic)
                        obs, reward, done, info = env.step(action)

                        episode_length += 1
                        episode_is_time_out.append(info[-1]["is_time_out"])
                        episode_is_collision.append(info[-1]["is_collision"])
                        episode_is_off_road.append(info[-1]["is_off_road"])
                        episode_is_goal_reached.append(info[-1]["is_goal_reached"])
                        episode_is_friction_violation.append(info[-1]["is_friction_violation"])

                        if callback is not None:
                            callback(locals(), globals())
                        if render:
                            env.render()

                    # Calculate cost for optimization from state information
                    normalized_episode_length = episode_length / info[-1]["max_episode_time_steps"]
                    if episode_is_time_out[-1]:
                        episode_cost += 10.0  # * (1 / normalized_episode_length)
                    if episode_is_collision[-1]:
                        episode_cost += 10.0  # * (1 / normalized_episode_length)
                    if episode_is_off_road[-1]:
                        episode_cost += 10.0  # * (1 / normalized_episode_length)
                    if episode_is_friction_violation[-1]:
                        episode_cost += (
                                10.0 * episode_is_friction_violation[-1] / episode_length
                        )  # * (1 / normalized_episode_length)
                    if episode_is_goal_reached[-1]:
                        episode_cost -= 10.0  # * normalized_episode_length

                    episode_costs.append(episode_cost)
                    episode_lengths.append(episode_length)

                return episode_costs, episode_lengths

            sync_envs_normalization(self.training_env, self.eval_env)
            episode_costs, episode_lengths = evaluate_policy_configs(
                self.model,
                self.eval_env,
                n_eval_episode=self.n_eval_episodes,
                render=self.render,
                deterministic=self.deterministic,
            )

            mean_cost, std_cost = np.mean(episode_costs), np.std(episode_costs)
            mean_length, std_length = np.mean(episode_lengths), np.std(episode_lengths)
            self.last_mean_cost = mean_cost

            LOGGER.info("Evaluating at learning time step: {}".format(self.num_timesteps))
            LOGGER.info("Cost mean: {:.2f}, std: {:.2f}".format(mean_cost, std_cost))
            LOGGER.info("Length mean: {:.2f}, std: {:.2f}".format(mean_length, std_length))

            if self.log_path is not None:
                self.evaluation_timesteps.append(self.num_timesteps)
                self.evaluation_costs.append(episode_costs)
                self.evaluation_lengths.append(episode_lengths)
                np.savez(
                    self.log_path,
                    timesteps=self.evaluation_timesteps,
                    episode_costs=self.evaluation_costs,
                    episode_lengths=self.evaluation_lengths,
                )

            if mean_cost < self.lowest_mean_cost:
                self.lowest_mean_cost = mean_cost
                if self.best_model_save_path is not None:
                    self.model.save(self.best_model_save_path)
                # Trigger callback if needed
                if self.callback is not None:
                    return self._on_event()

            self.eval_idx += 1
            self.cost = self.lowest_mean_cost
            self.trial.report(self.cost, self.eval_idx)
            # Prune trial if need
            if self.trial.should_prune():
                self.is_pruned = True
                return False
        return True


class MultiEnvsEvalCallback(EventCallback):
    """
    Callback for evaluating an agent with multiple environments, i.e. SubprocVecEnv
    """

    def __init__(
            self,
            eval_env: Union[gym.Env, VecEnv],
            callback_on_new_best: Optional[BaseCallback] = None,
            log_path: str = None,
            best_model_save_path: str = None,
            eval_freq: int = 1000,
            n_eval_timesteps: int = 1000,
            deterministic: bool = True,
            render: bool = False,
            verbose: int = 2,
    ):
        # Set basics
        super(MultiEnvsEvalCallback, self).__init__(callback_on_new_best, verbose=verbose)
        self.n_eval_timesteps = n_eval_timesteps
        self.eval_freq = eval_freq
        self.best_mean_reward = -np.inf
        self.last_mean_reward = -np.inf
        self.deterministic = deterministic
        self.render = render
        self.evaluations_timesteps = []
        self.evaluations_results = []
        self.evaluations_lengths = []

        # Convert to VecEnv for consistency
        if not isinstance(eval_env, VecEnv):
            eval_env = DummyVecEnv([lambda: eval_env])
        self.eval_env = eval_env

        # Store best model in `best_model.zip`
        self.best_model_save_path = construct_path_with_trial_number(best_model_save_path, "best_model")
        # Write logs in `evaluations.npz`
        self.log_path = construct_path_with_trial_number(log_path, "evaluations")

        # Set save path and logger
        if not len(LOGGER.handlers):
            map_verbose_to_logging = {2: logging.DEBUG, 1: logging.INFO}
            construct_logger(map_verbose_to_logging.get(verbose, logging.ERROR), log_path, LOGGER)

    def _init_callback(self):
        # Does not work in some corner cases, where the wrapper is not the same
        if not type(self.training_env) is type(self.eval_env):
            warnings.warn(f"Training and eval env are not of the same type {self.training_env} != {self.eval_env}")

    def _on_step(self) -> bool:

        if self.eval_freq > 0 and self.n_calls % self.eval_freq == 0:

            def evaluate_policy_multi_envs(
                    model: "BaseRLModel",
                    env: Union[gym.Env, VecEnv],
                    n_eval_timesteps: int = 1000,
                    deterministic: bool = True,
                    render: bool = False,
                    callback: Optional[Callable] = None,
            ) -> Tuple[List[float], List[int]]:
                """
                Runs policy on `env.num_envs` environments until `n_eval_timesteps` timesteps have been collected asynchronously,,
                and returns lists of timesteps rewards and timesteps lengths.
                """

                episode_rewards = []
                episode_lengths = []

                # Reset observation
                # Pad observation for recrrent policies
                # See https://github.com/hill-a/stable-baselines/issues/1015
                obs = env.reset()
                is_recurrent = model.policy.recurrent
                if is_recurrent:
                    zero_completed_obs = np.zeros((model.n_envs,) + model.observation_space.shape)
                    zero_completed_obs[0, :] = obs
                    obs = zero_completed_obs

                dones, states = [False] * env.num_envs, [None] * env.num_envs
                episode_rewards_tmp = [0.0] * env.num_envs
                episode_lengths_tmp = [0] * env.num_envs
                while True:
                    # Predict actions and take a step
                    actions, states = model.predict(obs, state=states, deterministic=deterministic)
                    new_obs, rewards, dones, _info = env.step(actions)

                    if is_recurrent:
                        obs[0, :] = new_obs
                    else:
                        obs = new_obs

                    # Collect results for each evaluating environment
                    for i in range(env.num_envs):
                        if dones[i]:
                            episode_rewards_tmp[i] += rewards[i]
                            episode_lengths_tmp[i] += 1
                            episode_rewards.append(episode_rewards_tmp[i])
                            episode_lengths.append(episode_lengths_tmp[i])
                            episode_rewards_tmp[i] = 0.0
                            episode_lengths_tmp[i] = 0
                            LOGGER.debug(f"Collected {len(episode_rewards)} evaluation episodes")
                        else:
                            episode_rewards_tmp[i] += rewards[i]
                            episode_lengths_tmp[i] += 1

                    if callback is not None:
                        callback(locals(), globals())
                    if render:
                        env.render()

                    # Break if n_eval_timesteps enough results are obtained
                    if sum(episode_lengths) >= n_eval_timesteps:
                        break

                return episode_rewards, episode_lengths

            # Sync training and evaluating envs if there is VecNormalize
            sync_envs_normalization(self.training_env, self.eval_env)

            episode_rewards, episode_lengths = evaluate_policy_multi_envs(self.model,
                                                                          self.eval_env,
                                                                          self.n_eval_timesteps,
                                                                          self.deterministic,
                                                                          self.render)

            if self.log_path is not None:
                self.evaluations_timesteps.append(self.num_timesteps)
                self.evaluations_results.append(episode_rewards)
                self.evaluations_lengths.append(episode_lengths)
                np.savez(self.log_path, timesteps=self.evaluations_timesteps,
                         results=self.evaluations_results, ep_lengths=self.evaluations_lengths)

            mean_reward, std_reward = np.mean(episode_rewards), np.std(episode_rewards)
            mean_ep_length, std_ep_length = np.mean(episode_lengths), np.std(episode_lengths)

            # Keep track of the last evaluation, useful for classes that derive from this callback
            self.last_mean_reward = mean_reward

            LOGGER.info("Eval num_timesteps={}, episode_reward={:.2f} +/- {:.2f}".format(
                self.num_timesteps, mean_reward, std_reward))
            LOGGER.info("Episode length: {:.2f} +/- {:.2f}".format(mean_ep_length, std_ep_length))

            if mean_reward > self.best_mean_reward:
                LOGGER.info("New best mean reward!")
                self.best_mean_reward = mean_reward
                if self.best_model_save_path is not None:
                    self.model.save(self.best_model_save_path)
                if self.callback is not None:
                    return self._on_event()

        return True


class SaveVecNormalizeCallback(BaseCallback):
    """
    Callback for saving a VecNormalize wrapper every ``save_freq`` steps

    :param save_freq: (int)
    :param save_path: (str) Path to the folder where ``VecNormalize`` will be saved, as ``vecnormalize.pkl``
    :param name_prefix: (str) Common prefix to the saved ``VecNormalize``, if None (default)
        only one file will be kept.
    """

    def __init__(self, save_freq: int, save_path: str, name_prefix=None, verbose=1):
        super(SaveVecNormalizeCallback, self).__init__(verbose)
        self.save_freq = save_freq
        self.save_path = save_path
        self.name_prefix = name_prefix

        # Set save path and logger
        if not len(LOGGER.handlers):
            map_verbose_to_logging = {2: logging.DEBUG, 1: logging.INFO}
            construct_logger(map_verbose_to_logging.get(verbose, logging.ERROR), save_path, LOGGER)

    def _init_callback(self) -> None:
        # Create folder if needed
        if self.save_path is not None:
            os.makedirs(self.save_path, exist_ok=True)

    def _on_step(self) -> bool:
        if self.n_calls % self.save_freq == 0:
            if self.name_prefix is not None:
                path = os.path.join(self.save_path, f"{self.name_prefix}_{self.num_timesteps}_steps.pkl")
            else:
                path = os.path.join(self.save_path, "vecnormalize.pkl")

            LOGGER.info(f"Saving vectorized and normalized environment wrapper to {self.save_path}")
            self.model.get_vec_normalize_env().save(path)
        return True



"""Adapted from https://github.com/openai/baselines/blob/master/baselines/common/plot_util.py"""
import json
import os
import os.path as osp
from collections import defaultdict, namedtuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas
from stable_baselines.bench import monitor
from stable_baselines.logger import read_json, read_csv


def smooth(y, radius, mode='two_sided', valid_only=False):
    '''
    Smooth signal y, where radius is determines the size of the window

    mode='twosided':
        average over the window [max(index - radius, 0), min(index + radius, len(y)-1)]
    mode='causal':
        average over the window [max(index - radius, 0), index]

    valid_only: put nan in entries where the full-sized window is not available

    '''
    assert mode in ('two_sided', 'causal')
    if len(y) < 2 * radius + 1:
        return np.ones_like(y) * y.mean()
    elif mode == 'two_sided':
        convkernel = np.ones(2 * radius + 1)
        out = np.convolve(y, convkernel, mode='same') / np.convolve(np.ones_like(y), convkernel, mode='same')
        if valid_only:
            out[:radius] = out[-radius:] = np.nan
    elif mode == 'causal':
        convkernel = np.ones(radius)
        out = np.convolve(y, convkernel, mode='full') / np.convolve(np.ones_like(y), convkernel, mode='full')
        out = out[:-radius + 1]
        if valid_only:
            out[:radius] = np.nan
    return out


def one_sided_ema(xolds, yolds, low=None, high=None, n=512, decay_steps=1., low_counts_threshold=1e-8):
    '''
    perform one-sided (causal) EMA (exponential moving average)
    smoothing and resampling to an even grid with n points.
    Does not do extrapolation, so we assume
    xolds[0] <= low && high <= xolds[-1]

    Arguments:

    xolds: array or list  - x values of data. Needs to be sorted in ascending order
    yolds: array of list  - y values of data. Has to have the same length as xolds

    low: float            - min value of the new x grid. By default equals to xolds[0]
    high: float           - max value of the new x grid. By default equals to xolds[-1]

    n: int                - number of points in new x grid

    decay_steps: float    - EMA decay factor, expressed in new x grid steps.

    low_counts_threshold: float or int
                          - y values with counts less than this value will be set to NaN

    Returns:
        tuple sum_ys, count_ys where
            xs        - array with new x grid
            ys        - array of EMA of y at each point of the new x grid
            count_ys  - array of EMA of y counts at each point of the new x grid

    '''

    low = xolds[0] if low is None else low
    high = xolds[-1] if high is None else high

    assert xolds[0] <= low, 'low = {} < xolds[0] = {} - extrapolation not permitted!'.format(low, xolds[0])
    assert xolds[-1] >= high, 'high = {} > xolds[-1] = {}  - extrapolation not permitted!'.format(high, xolds[-1])
    assert len(xolds) == len(yolds), 'length of xolds ({}) and yolds ({}) do not match!'.format(len(xolds), len(yolds))

    xolds = xolds.astype('float64')
    yolds = yolds.astype('float64')

    luoi = 0  # last unused old index
    sum_y = 0.
    count_y = 0.
    xnews = np.linspace(low, high, n)
    decay_period = (high - low) / (n - 1) * decay_steps
    interstep_decay = np.exp(- 1. / decay_steps)
    sum_ys = np.zeros_like(xnews)
    count_ys = np.zeros_like(xnews)
    for i in range(n):
        xnew = xnews[i]
        sum_y *= interstep_decay
        count_y *= interstep_decay
        while True:
            if luoi >= len(xolds):
                break
            xold = xolds[luoi]
            if xold <= xnew:
                decay = np.exp(- (xnew - xold) / decay_period)
                sum_y += decay * yolds[luoi]
                count_y += decay
                luoi += 1
            else:
                break
        sum_ys[i] = sum_y
        count_ys[i] = count_y

    ys = sum_ys / count_ys
    ys[count_ys < low_counts_threshold] = np.nan

    return xnews, ys, count_ys


def symmetric_ema(xolds, yolds, low=None, high=None, n=512, decay_steps=1., low_counts_threshold=1e-8):
    '''
    perform symmetric EMA (exponential moving average)
    smoothing and resampling to an even grid with n points.
    Does not do extrapolation, so we assume
    xolds[0] <= low && high <= xolds[-1]

    Arguments:

    xolds: array or list  - x values of data. Needs to be sorted in ascending order
    yolds: array of list  - y values of data. Has to have the same length as xolds

    low: float            - min value of the new x grid. By default equals to xolds[0]
    high: float           - max value of the new x grid. By default equals to xolds[-1]

    n: int                - number of points in new x grid

    decay_steps: float    - EMA decay factor, expressed in new x grid steps.

    low_counts_threshold: float or int
                          - y values with counts less than this value will be set to NaN

    Returns:
        tuple sum_ys, count_ys where
            xs        - array with new x grid
            ys        - array of EMA of y at each point of the new x grid
            count_ys  - array of EMA of y counts at each point of the new x grid

    '''
    xs, ys1, count_ys1 = one_sided_ema(xolds, yolds, low, high, n, decay_steps, low_counts_threshold=0)
    _, ys2, count_ys2 = one_sided_ema(-xolds[::-1], yolds[::-1], -high, -low, n, decay_steps, low_counts_threshold=0)
    ys2 = ys2[::-1]
    count_ys2 = count_ys2[::-1]
    count_ys = count_ys1 + count_ys2
    ys = (ys1 * count_ys1 + ys2 * count_ys2) / count_ys
    ys[count_ys < low_counts_threshold] = np.nan
    return xs, ys, count_ys


Result = namedtuple('Result', 'monitor progress dirname metadata')
Result.__new__.__defaults__ = (None,) * len(Result._fields)


def load_results(root_dir_or_dirs, enable_progress=True, enable_monitor=True, verbose=False):
    '''
    load summaries of runs from a list of directories (including subdirectories)
    Arguments:

    enable_progress: bool - if True, will attempt to load data from progress.csv files (data saved by logger).
    Default: True

    enable_monitor: bool - if True, will attempt to load data from monitor.csv files (data saved by Monitor
    environment wrapper). Default: True

    verbose: bool - if True, will print out list of directories from which the data is loaded. Default: False


    Returns:
    List of Result objects with the following fields:
         - dirname - path to the directory data was loaded from
         - metadata - run metadata (such as command-line arguments and anything else in metadata.json file
         - monitor - if enable_monitor is True, this field contains pandas dataframe with loaded monitor.csv file (or
         aggregate of all *.monitor.csv files in the directory)
         - progress - if enable_progress is True, this field contains pandas dataframe with loaded progress.csv file
    '''
    import re
    if isinstance(root_dir_or_dirs, str):
        rootdirs = [osp.expanduser(root_dir_or_dirs)]
    else:
        rootdirs = [osp.expanduser(d) for d in root_dir_or_dirs]
    allresults = []
    for rootdir in sorted(rootdirs):
        assert osp.exists(rootdir), "%s doesn't exist" % rootdir
        for dirname, dirs, files in os.walk(rootdir):
            if '-proc' in dirname:
                files[:] = []
                continue
            monitor_re = re.compile(r'(\d+\.)?(\d+\.)?monitor\.csv')
            if set(['metadata.json', 'monitor.json', 'progress.json', 'progress.csv']).intersection(files) or any(
                    [f for f in files if monitor_re.match(f)]):  # also match monitor files like 0.1.monitor.csv
                # used to be uncommented, which means do not go deeper than current directory if any of the data files
                # are found
                # dirs[:] = []
                result = {'dirname': dirname}
                if "metadata.json" in files:
                    with open(osp.join(dirname, "metadata.json"), "r") as fh:
                        result['metadata'] = json.load(fh)
                progjson = osp.join(dirname, "progress.json")
                progcsv = osp.join(dirname, "progress.csv")
                if enable_progress:
                    if osp.exists(progjson):
                        result['progress'] = pandas.DataFrame(read_json(progjson))
                    elif osp.exists(progcsv):
                        try:
                            result['progress'] = read_csv(progcsv)
                        except pandas.errors.EmptyDataError:
                            print('skipping progress file in ', dirname, 'empty data')
                    else:
                        if verbose: print('skipping %s: no progress file' % dirname)

                if enable_monitor:
                    try:
                        monitor_result = monitor.load_results(dirname)
                        result['monitor'] = pandas.DataFrame(monitor_result)
                    except monitor.LoadMonitorResultsError:
                        print('skipping %s: no monitor files' % dirname)
                    except Exception as e:
                        print('exception loading monitor file in %s: %s' % (dirname, e))

                if result.get('monitor') is not None or result.get('progress') is not None:
                    allresults.append(Result(**result))
                    if verbose:
                        print('successfully loaded %s' % dirname)

    if verbose: print('loaded %i results' % len(allresults))
    return allresults


COLORS = ["tab:red", "cornflowerblue", "slategray", "darkorange", 'royalblue', 'lightcoral', "tab:pink", "tab:red",
          'blue', 'green', 'red', 'cyan', 'magenta', 'yellow', 'black', 'purple', 'pink', 'brown', 'orange', 'teal',
          'lightblue', 'lime', 'lavender', 'turquoise', 'darkgreen', 'tan', 'salmon', 'gold', 'darkred', 'darkblue']  #


def default_xy_fn(r):
    x = np.cumsum(r.monitor.l)
    y = smooth(r.monitor.r, radius=10)
    return x, y


def default_split_fn(r):
    import re
    # match name between slash and -<digits> at the end of the string
    # (slash in the beginning or -<digits> in the end or either may be missing)
    match = re.search(r'[^/-]+(?=(_\d+)?\Z)', r.dirname)
    if match:
        return match.group(0)


def plot_results(
        allresults, f, axarr, *, xy_fn=default_xy_fn, split_fn=default_split_fn, group_fn=default_split_fn,
        average_group=False, shaded_std=True, shaded_err=True, figsize=None, legend_outside=False, resample=0,
        smooth_step=1.0, nrows=1, ncols=1, idx_row=0, idx_col=0, xlabel=None, ylabel=None, label=None, labelpad=None,
        legend=True, plot_line=False, set_y_lim=False):
    '''
    Plot multiple Results objects

    xy_fn: function Result -> x,y           - function that converts results objects into tuple of x and y values.
                                              By default, x is cumsum of episode lengths, and y is episode rewards

    split_fn: function Result -> hashable   - function that converts results objects into keys to split curves into
    sub-panels by.
                                              That is, the results r for which split_fn(r) is different will be put
                                              on different sub-panels.
                                              By default, the portion of r.dirname between last / and -<digits> is
                                              returned. The sub-panels are
                                              stacked vertically in the figure.

    group_fn: function Result -> hashable   - function that converts results objects into keys to group curves by.
                                              That is, the results r for which group_fn(r) is the same will be put
                                              into the same group.
                                              Curves in the same group have the same color (if average_group is
                                              False), or averaged over
                                              (if average_group is True). The default value is the same as default
                                              value for split_fn

    average_group: bool                     - if True, will average the curves in the same group and plot the mean.
    Enables resampling
                                              (if resample = 0, will use 512 steps)

    shaded_std: bool                        - if True (default), the shaded region corresponding to standard
    deviation of the group of curves will be
                                              shown (only applicable if average_group = True)

    shaded_err: bool                        - if True (default), the shaded region corresponding to error in mean
    estimate of the group of curves
                                              (that is, standard deviation divided by square root of number of
                                              curves) will be
                                              shown (only applicable if average_group = True)

    figsize: tuple or None                  - size of the resulting figure (including sub-panels). By default,
    width is 6 and height is 6 times number of
                                              sub-panels.


    legend_outside: bool                    - if True, will place the legend outside of the sub-panels.

    resample: int                           - if not zero, size of the uniform grid in x direction to resample onto.
    Resampling is performed via symmetric
                                              EMA smoothing (see the docstring for symmetric_ema).
                                              Default is zero (no resampling). Note that if average_group is True,
                                              resampling is necessary; in that case, default
                                              value is 512.

    smooth_step: float                      - when resampling (i.e. when resample > 0 or average_group is True),
    use this EMA decay parameter (in units of the new grid step).
                                              See docstrings for decay_steps in symmetric_ema or one_sided_ema
                                              functions.

    '''

    if split_fn is None: split_fn = lambda _: ''
    if group_fn is None: group_fn = lambda _: ''
    sk2r = defaultdict(list)  # splitkey2results
    for result in allresults:
        splitkey = split_fn(result)
        sk2r[splitkey].append(result)
    assert len(sk2r) > 0
    assert isinstance(resample, int), "0: don't resample. <integer>: that many samples"

    groups = list(set(group_fn(result) for result in allresults))

    default_samples = 512
    if average_group:
        resample = resample or default_samples

    for (isplit, sk) in enumerate(sorted(sk2r.keys())):
        g2l = {}
        g2c = defaultdict(int)
        sresults = sk2r[sk]
        gresults = defaultdict(list)
        # idx_row = isplit // ncols
        # idx_col = isplit % ncols
        ax = axarr[idx_row][idx_col]
        for result in sresults:
            group = group_fn(result)
            g2c[group] += 1
            x, y = xy_fn(result)
            if x is None: x = np.arange(len(y))
            x, y = map(np.asarray, (x, y))

            if average_group:
                gresults[group].append((x, y))
            else:
                color = COLORS[groups.index(group) % len(COLORS)]
                if resample:
                    # if len(x) < 10000:
                    #     smooth_window = 100
                    # elif len(x) < int(1e5):
                    #     smooth_window = 1000
                    # else:
                    #     smooth_window = 10000
                    smooth_window = 100
                    y = smooth(y, radius=smooth_window)
                    ymean, ystd, ystderr = [], [], []
                    xs = np.array([xx for xx in x[smooth_window:-smooth_window]])
                    for i in range(smooth_window, len(y) - smooth_window):
                        ys = y[i - smooth_window:i + smooth_window]
                        ymean.append(np.mean(ys))
                        ystd.append(np.std(ys))
                        ystderr.append(np.max(np.abs(ys)))
                    ymean = np.array(ymean)
                    ystd = np.array(ystd)
                    ystderr = ystd / np.sqrt(len(y))  # np.array(ystderr)
                    y = ymean
                    x = xs

                    # if os.path.basename(result.dirname) == "safe agent":
                    #     x = np.concatenate((np.array([0.09548599999999999]), x))
                    #     y = np.concatenate((np.array([y[0]]), y))
                    # show shaded std
                    if shaded_std:
                        ax.fill_between(x, ymean - ystd, ymean + ystd, color=color, alpha=.2, rasterized=True)
                group = group.replace("_", " ")
                if label is not None:
                    l, = ax.plot(x, y, color=color, label=label)
                else:
                    l, = ax.plot(x, y, color=color, label=group) # label="\\textbf{" + group + "}")
                if plot_line:
                    ax.plot([x[0], x[-1]], [0., 0.], "--", color="lightgray")
                g2l[group] = l
        if average_group:
            for group in sorted(groups):
                xys = gresults[group]
                if not any(xys):
                    continue
                color = COLORS[groups.index(group) % len(COLORS)]
                origxs = [xy[0] for xy in xys]
                minxlen = min(map(len, origxs))

                def allequal(qs):
                    return all((q == qs[0]).all() for q in qs[1:])

                if resample:
                    low = max(x[0] for x in origxs)
                    high = min(x[-1] for x in origxs)
                    usex = np.linspace(low, high, resample)
                    ys = []
                    for (x, y) in xys:
                        ys.append(symmetric_ema(x, y, low, high, resample, decay_steps=smooth_step)[1])
                else:
                    assert allequal([x[:minxlen] for x in origxs]), \
                        'If you want to average unevenly sampled data, set resample=<number of samples you want>'
                    usex = origxs[0]
                    ys = [xy[1][:minxlen] for xy in xys]
                ymean = np.mean(ys, axis=0)
                ystd = np.std(ys, axis=0)
                ystderr = ystd / np.sqrt(len(ys))
                l, = axarr[idx_row][idx_col].plot(usex, ymean, color=color)
                g2l[group] = l
                if shaded_err:
                    ax.fill_between(usex, ymean - ystderr, ymean + ystderr, color=color, alpha=.4)
                if shaded_std:
                    ax.fill_between(usex, ymean - ystd, ymean + ystd, color=color, alpha=.2)

        # ax.set_xlim([np.min(x) - 1, np.max(x)])
#        ax.set_xlim([np.min(x)-0.005, np.max(x)])
#        print(np.min(x)-0.005)
        if set_y_lim:
            ax.set_ylim([-0.05, 1.05])

        # locs = [int(xtick) for xtick in list(ax.get_xticks())] + [int(np.max(x))]
        locs = list(ax.get_xticks()) + [1000]
        locator = matplotlib.ticker.FixedLocator(locs)
        ax.xaxis.set_major_locator(locator)
        # https://matplotlib.org/users/legend_guide.html
        # plt.tight_layout()
        if legend:
            if label is not None:
                ax.legend(label)
            elif any(g2l.keys()):
                s = f.subplotpars
                # bb = [s.left+0.2, s.top+0.02, s.right-s.left - 0.6, 0.05] #
                bb = [s.left + 0.05, s.top + 0.25, s.right - s.left + 1.0, 0.05]
                # leg = axes[0].legend(loc=8, bbox_to_anchor=bb, ncol=4, mode="expand", borderaxespad=0,
                #                      bbox_transform=fig.transFigure, fancybox=False, edgecolor="k")
                # ax.legend(loc=8, bbox_to_anchor=bb, ncol=2, mode="expand", borderaxespad=0,
                # bbox_transform=f.transFigure, fancybox=False, edgecolor="k")
                ax.legend(ncol=1, #len(groups),
                          # g2l.values(),
                          # ['%s (%i)'%(g, g2c[g]) for g in g2l] if average_group else g2l.keys(),
                          loc=8 if legend_outside else None, bbox_to_anchor=bb if legend_outside else None,
                          mode="expand", borderaxespad=0, fancybox=False)

        # ax.set_title(sk)
        # add xlabels, but only to the bottom row
        if xlabel is not None:
            plt.sca(axarr[idx_row][idx_col])
            plt.xlabel(xlabel)  # for ax in axarr[-1]:  #     plt.sca(ax)  #     plt.xlabel(xlabel)
        # add ylabels, but only to left column
        if ylabel is not None:
            plt.sca(axarr[idx_row][idx_col])
            if labelpad is not None:
                plt.ylabel(ylabel, labelpad=labelpad)
            else:
                plt.ylabel(ylabel)  # for ax in axarr: #[:, 0]:  #     plt.sca(ax)  #     plt.ylabel(ylabel)

    return f, axarr


def test_smooth():
    norig = 100
    nup = 300
    ndown = 30
    xs = np.cumsum(np.random.rand(norig) * 10 / norig)
    yclean = np.sin(xs)
    ys = yclean + .1 * np.random.randn(yclean.size)
    xup, yup, _ = symmetric_ema(xs, ys, xs.min(), xs.max(), nup, decay_steps=nup / ndown)
    xdown, ydown, _ = symmetric_ema(xs, ys, xs.min(), xs.max(), ndown, decay_steps=ndown / ndown)
    xsame, ysame, _ = symmetric_ema(xs, ys, xs.min(), xs.max(), norig, decay_steps=norig / ndown)
    plt.plot(xs, ys, label='orig', marker='x')
    plt.plot(xup, yup, label='up', marker='x')
    plt.plot(xdown, ydown, label='down', marker='x')
    plt.plot(xsame, ysame, label='same', marker='x')
    plt.plot(xs, yclean, label='clean', marker='x')
    plt.legend()
    plt.show()



"""
Module containing the action base class
"""
import gym
from typing import Union
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem
from commonroad_rl.gym_commonroad.action.vehicle import *


def _rotate_to_curvi(vector: np.ndarray, local_ccosy: CurvilinearCoordinateSystem, pos: np.ndarray) \
        -> np.ndarray:
    """
    Function to rotate a vector in the curvilinear system to its counterpart in the normal coordinate system

    :param vector: The vector in question
    :returns: The rotated vector
    """
    try:
        long, _ = local_ccosy.convert_to_curvilinear_coords(pos[0], pos[1])
    except ValueError:
        long = 0.

    tangent = local_ccosy.tangent(long)
    theta = np.math.atan2(tangent[1], tangent[0])
    rot_mat = np.array([[np.cos(theta), -np.sin(theta)],
                        [np.sin(theta), np.cos(theta)]])

    return np.matmul(rot_mat, vector)


class Action(ABC):
    """
    Description:
        Abstract base class of all action spaces
    """

    def __init__(self):
        """ Initialize empty object """
        super().__init__()
        self.vehicle = None

    @abstractmethod
    def step(self, action: Union[np.ndarray, int], local_ccosy: CurvilinearCoordinateSystem = None) -> None:
        """
        Function which acts on the current state and generates the new state
        :param action: current action
        :param local_ccosy: Current curvilinear coordinate system
        """
        pass


class DiscreteAction(Action):
    """
    Description:
        Abstract base class of all discrete action spaces. Each high-level discrete
        action is converted to a low-level trajectory by a specified planner.
    """

    def __init__(self, vehicle_params_dict: dict, long_steps: int, lat_steps: int):
        """ Initialize empty object """
        super().__init__()

        assert VehicleModel(vehicle_params_dict["vehicle_model"]) == VehicleModel.PM, \
            'ERROR in ACTION INITIALIZATION: DiscreteAction only supports the PM vehicle_type no'

        assert long_steps % 2 != 0 and lat_steps % 2 != 0, \
            'ERROR in ACTION INITIALIZATION: The discrete steps for longitudinal and lateral action ' \
            'have to be odd numbers, so constant velocity without turning is an possible action'

        self.vehicle = ContinuousVehicle(vehicle_params_dict)
        self.local_ccosy = None

    def reset(self, initial_state: State, dt: float) -> None:
        """
        resets the vehicle
        :param initial_state: initial state
        :param dt: time step size of scenario
        """
        self.vehicle.reset(initial_state, dt)

    def step(self, action: Union[np.ndarray, int], local_ccosy: CurvilinearCoordinateSystem = None) -> None:
        """
        Function which acts on the current state and generates the new state

        :param action: current action
        :param local_ccosy: Current curvilinear coordinate system
        """
        self.local_ccosy = local_ccosy
        state = self.get_new_state(action)
        self.vehicle.set_current_state(state)

    @abstractmethod
    def get_new_state(self, action: Union[np.ndarray, int]) -> State:
        """function which return new states given the action and current state"""
        pass

    def _propagate(self, control_input: np.array):
        # Rotate the action according to the curvilinear coordinate system
        if self.local_ccosy is not None:
            control_input = _rotate_to_curvi(control_input, self.local_ccosy, self.vehicle.state.position)

        # get the next state from the PM model
        return self.vehicle.get_new_state(control_input, "acceleration")


class DiscretePMJerkAction(DiscreteAction):
    """
    Description:
        Discrete / High-level action class with point mass model and jerk control
    """

    def __init__(self, vehicle_params_dict: dict, long_steps: int, lat_steps: int):
        """
        Initialize object
        :param vehicle_params_dict: vehicle parameter dictionary
        :param long_steps: number of discrete longitudinal jerk steps
        :param lat_steps: number of discrete lateral jerk steps
        """
        super().__init__(vehicle_params_dict, long_steps, lat_steps)

        self.j_max = 10  # set the maximum jerk
        self.long_step_size = (self.j_max * 2) / (long_steps - 1)
        self.lat_step_size = (self.j_max * 2) / (lat_steps - 1)
        self.action_mapping_long = {}
        self.action_mapping_lat = {}

        for idx in range(long_steps):
            self.action_mapping_long[idx] = (self.j_max - (idx * self.long_step_size))

        for idx in range(lat_steps):
            self.action_mapping_lat[idx] = (self.j_max - (idx * self.lat_step_size))

    def get_new_state(self, action: Union[np.ndarray, int]) -> State:
        """
        calculation of next state depending on the discrete action
        :param action: discrete action
        :return: next state
        """
        # map discrete action to jerk and calculate a
        # correct rescale in order to make 0 acceleration achievable again when sign of acc switches
        a_long = self.action_mapping_long[action[0]] * self.vehicle.dt + self.vehicle.state.acceleration
        if self.vehicle.state.acceleration != 0 and np.sign(a_long) != np.sign(self.vehicle.state.acceleration) and \
                (np.abs(a_long) % (self.long_step_size * self.vehicle.dt)) != 0:
            a_long = self.action_mapping_long[action[0]] * self.vehicle.dt + self.vehicle.state.acceleration - \
                     np.sign(a_long) * (np.abs(a_long) % (self.long_step_size * self.vehicle.dt))

        a_lat = self.action_mapping_lat[action[1]] * self.vehicle.dt + self.vehicle.state.acceleration_y
        if self.vehicle.state.acceleration_y != 0 and np.sign(a_lat) != np.sign(self.vehicle.state.acceleration_y) and \
                (np.abs(a_lat) % (self.lat_step_size * self.vehicle.dt)) != 0:
            a_lat = self.action_mapping_long[action[1]] * self.vehicle.dt + self.vehicle.state.acceleration_y - \
                    np.sign(a_lat) * (np.abs(a_lat) % (self.lat_step_size * self.vehicle.dt))

        control_input = np.array([a_long, a_lat])

        return self._propagate(control_input)


class DiscretePMAction(DiscreteAction):
    """
    Description:
        Discrete / High-level action class with point mass model
    """

    def __init__(self, vehicle_params_dict: dict, long_steps: int, lat_steps: int):
        """
        Initialize object
        :param vehicle_params_dict: vehicle parameter dictionary
        :param long_steps: number of discrete acceleration steps
        :param lat_steps: number of discrete turning steps
        """
        super().__init__(vehicle_params_dict, long_steps, lat_steps)

        a_max = self.vehicle.parameters.longitudinal.a_max
        a_long_steps = (a_max * 2) / (long_steps - 1)
        a_lat_steps = (a_max * 2) / (lat_steps - 1)

        self.action_mapping_long = {}
        self.action_mapping_lat = {}

        for idx in range(long_steps):
            self.action_mapping_long[idx] = (a_max - (idx * a_long_steps))

        for idx in range(lat_steps):
            self.action_mapping_lat[idx] = (a_max - (idx * a_lat_steps))

    def propogate_one_state(self, state: State, action: Union[np.ndarray, int]):
        """
        Used to generate a trajectory from a given action
        :param state:
        :param action:
        :return:
        """
        control_input = np.array([self.action_mapping_long[action[0]],
                                  self.action_mapping_lat[action[1]]])
        # Rotate the action according to the curvilinear coordinate system
        if self.local_ccosy is not None:
            control_input = _rotate_to_curvi(control_input, self.local_ccosy, state.position)

        return self.vehicle.propagate_one_time_step(state, control_input, "acceleration")

    def get_new_state(self, action: Union[np.ndarray, int]) -> State:
        """
        calculation of next state depending on the discrete action
        :param action: discrete action
        :return: next state
        """
        return self.propogate_one_state(state=self.vehicle.state, action=action)


class ContinuousAction(Action):
    """
    Description:
        Module for continuous action space; actions correspond to vehicle control inputs
    """

    def __init__(self, params_dict: dict, action_dict: dict):
        """ Initialize object """
        super().__init__()
        # create vehicle object
        self.action_base = action_dict['action_base']
        self._continous_collision_check = action_dict.get("continuous_collision_checking", True)
        self.vehicle = ContinuousVehicle(params_dict, continuous_collision_checking=self._continous_collision_check)

    def _set_rescale_factors(self):

        a_max = self.vehicle.parameters.longitudinal.a_max
        # rescale factors for PM model
        if self.vehicle.vehicle_model == VehicleModel.PM:
            self._rescale_factor = np.array([a_max, a_max])
            self._rescale_bias = 0.0
        # rescale factors for KS model
        elif self.vehicle.vehicle_model == VehicleModel.KS:
            steering_v_max = self.vehicle.parameters.steering.v_max
            steering_v_min = self.vehicle.parameters.steering.v_min
            self._rescale_factor = np.array([(steering_v_max - steering_v_min) / 2., a_max])
            self._rescale_bias = np.array([(steering_v_max + steering_v_min) / 2., 0.])
        # rescale factors for YawRate model
        elif self.vehicle.vehicle_model == VehicleModel.YawRate:
            yaw_rate_max = self.vehicle.parameters.yaw.v_max = np.abs(
                self.vehicle.parameters.longitudinal.a_max / (self.vehicle.state.velocity + 1e-6))
            yaw_rate_min = self.vehicle.parameters.yaw.v_min = -self.vehicle.parameters.yaw.v_max

            self._rescale_factor = np.array([(yaw_rate_max - yaw_rate_min) / 2., a_max])
            self._rescale_bias = np.array([0., 0.])
        elif self.vehicle.vehicle_model == VehicleModel.QP:
            ub, lb = self.vehicle.vehicle_dynamic.input_bounds.ub, self.vehicle.vehicle_dynamic.input_bounds.lb
            self._rescale_factor = (ub - lb) / 2
            self._rescale_bias = (ub + lb) / 2
        elif self.vehicle.vehicle_model == VehicleModel.PMNonlinear:
            self._rescale_factor = np.array([a_max, 2.]) # TODO: check with Niklas which bounds are used in the reachable sets
            self._rescale_bias = np.array([0., 0.])
        else:
            raise ValueError(
                f"action.py/_set_rescale_factors: rescale factors not defined for model {self.vehicle.vehicle_model}")

    def reset(self, initial_state: State, dt: float) -> None:
        self.vehicle.reset(initial_state, dt)
        self._set_rescale_factors()

    def step(self, action: Union[np.ndarray, int], local_ccosy: CurvilinearCoordinateSystem = None) -> None:
        """
        Function which acts on the current state and generates the new state

        :param action: current action
        :param local_ccosy: Current curvilinear coordinate system
        :return: New state of ego vehicle
        """
        rescaled_action = self.rescale_action(action)
        new_state = self.vehicle.get_new_state(rescaled_action, self.action_base)
        self.vehicle.set_current_state(new_state)
        if self.vehicle.vehicle_model == VehicleModel.YawRate:
            self._set_rescale_factors()

    def rescale_action(self, action: np.ndarray) -> np.ndarray:
        """
        Rescales the normalized action from [-1,1] to the required range

        :param action: action from the CommonroadEnv.
        :return: rescaled action
        """
        assert hasattr(self, "_rescale_bias") and hasattr(self, "_rescale_factor"), \
            "<ContinuousAction/rescale_action>: rescale factors not set, please run action.reset() first"

        return self._rescale_factor * action + self._rescale_bias


def action_constructor(action_configs: dict, vehicle_params: dict) -> \
        Tuple[Action, Union[gym.spaces.Box, gym.spaces.MultiDiscrete]]:

    if action_configs['action_type'] == "continuous":
        action = ContinuousAction(vehicle_params, action_configs)
    elif action_configs['action_type'] == "discrete":
        if action_configs['action_base'] == "acceleration":
            action = DiscretePMAction
        elif action_configs['action_base'] == "jerk":
            action = DiscretePMJerkAction
        else:
            raise NotImplementedError(f"action_base {action_configs['action_base']} not supported. "
                                      f"Please choose acceleration or jerk")
        action = action(vehicle_params, action_configs['long_steps'], action_configs['lat_steps'])
    else:
        raise NotImplementedError(f"action_type {action_configs['action_type']} not supported. "
                                  f"Please choose continuous or discrete")

        # Action space remove
        # TODO initialize action space with class
    if action_configs['action_type'] == "continuous":
        action_high = np.array([1.0, 1.0])
        action_space = gym.spaces.Box(low=-action_high, high=action_high, dtype="float32")
    else:
        action_space = gym.spaces.MultiDiscrete([action_configs['long_steps'], action_configs['lat_steps']])

    return action, action_space


"""Unit tests for termination class"""

from commonroad_rl.gym_commonroad.reward.termination import Termination
from commonroad_rl.tests.common.marker import unit_test, functional

configs = {
    "termination_configs": {
        "terminate_on_off_road": True
    }
}


@unit_test
@functional
def test_termination():
    termination = Termination(configs)
    observation = {
        "is_goal_reached": [False],
        "is_collision": [False],
        "is_off_road": [False],
        "is_time_out": [False],
        "is_friction_violation": [False],
    }

    done, _, _ = termination.is_terminated(observation, None)
    assert not done

    observation = {
        "is_goal_reached": [False],
        "is_collision": [False],
        "is_off_road": [True],
        "is_time_out": [False],
        "is_friction_violation": [False],
    }

    done, reason, _ = termination.is_terminated(observation, None)
    assert done
    assert reason == "is_off_road"



import pickle
from commonroad.common.util import Interval, AngleInterval
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblem
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario, ScenarioID
from commonroad.geometry.shape import Rectangle
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object
from commonroad_route_planner.route_planner import RoutePlanner
from shapely.geometry import LineString, Point

from commonroad_rl.gym_commonroad.action.action import *
from commonroad_rl.gym_commonroad.observation import LaneletNetworkObservation, ObservationCollector
from commonroad_rl.gym_commonroad.utils.scenario_io import restore_scenario
from commonroad_rl.gym_commonroad.utils.scenario import parse_map_name
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.path import *

xml_path = os.path.join(resource_root("test_commonroad_env"))
pickle_path = os.path.join(resource_root("test_commonroad_env"), "pickles")


def prepare_for_test():
    pickle_xml_scenarios(
        input_dir=xml_path,
        output_dir=pickle_path
    )
    # Lanelet observation settings
    lanelet_observation = LaneletNetworkObservation(
        configs={"lanelet_configs":
                     {"observe_left_marker_distance": True,
                      "observe_right_marker_distance": True,
                      "observe_left_road_edge_distance": True,
                      "observe_right_road_edge_distance": True,
                      "observe_is_off_road": True,
                      "strict_off_road_check": True,
                      "observe_lat_offset": True
                      }}
    )
    # construct ego vehicle
    vehicle_type = VehicleType.BMW_320i
    vehicle_model = VehicleModel.KS
    vehicle = ContinuousAction({"vehicle_type": vehicle_type, "vehicle_model": vehicle_model},
                               {"action_base": "acceleration"})

    # Initial state of the vehicle
    dummy_state = {
        "position": np.array([0.0, 0.0]),
        "velocity": 30.0,
        "yaw_rate": 0.0,
        "steering_angle": 0.0,
        "time_step": 0,
        "orientation": 0.0,
    }

    # Load meta scenario and problem dict
    filename = "DEU_A9-2_1_T-1.pickle"

    meta_scenario_reset_dict_path = os.path.join(pickle_path, "meta_scenario", "meta_scenario_reset_dict.pickle")
    with open(meta_scenario_reset_dict_path, "rb") as f:
        meta_scenario_reset_dict = pickle.load(f)

    # Load scenarios and problems
    fn = os.path.join(pickle_path, "problem", filename)
    with open(fn, "rb") as f:
        problem_dict = pickle.load(f)

    # Set scenario and problem
    scenario_id = ScenarioID.from_benchmark_id(os.path.basename(fn).split(".")[0], "2020a")
    map_id = parse_map_name(scenario_id)
    reset_config = meta_scenario_reset_dict[map_id]
    scenario = restore_scenario(reset_config["meta_scenario"], problem_dict["obstacle"], scenario_id)

    # specify laneletpolygons
    lanelet_polygons = [(l.lanelet_id, l.convert_to_polygon()) for l in scenario.lanelet_network.lanelets]
    lanelet_polygons_sg = pycrcc.ShapeGroup()
    for l_id, poly in lanelet_polygons:
        lanelet_polygons_sg.add_shape(create_collision_object(poly))
    dummy_time_step = Interval(0.0, 0.0)
    return lanelet_observation, vehicle, dummy_state, scenario, reset_config, \
           lanelet_polygons, lanelet_polygons_sg, dummy_time_step


lanelet_observation, vehicle_action, dummy_state, scenario, road_edge, lanelet_polygons, \
lanelet_polygons_sg, dummy_time_step = prepare_for_test()


@pytest.mark.parametrize(
    ('goal_region', 'ego_state', 'sampling_points', 'static', 'expected_offset'),
    [
        (
                GoalRegion([CustomState(time_step=Interval(1, 1), orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([3.0, 3.0])))]),
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": np.pi / 2,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 10
                }),
                [1],
                False,
                np.array([-6.])
        ),
        (
                GoalRegion([CustomState(time_step=Interval(1, 1), orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([0.0, 0.0])))]),
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": 0,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 4
                }),
                [1],
                True,
                np.array([0])
        ),
        (
                GoalRegion([CustomState(time_step=Interval(1, 1), orientation=AngleInterval(-np.pi / 2, np.pi / 2),
                                  position=Rectangle(length=2.0, width=2.0, center=np.array([3.0, 3.0])))]),
                CustomState(**{
                    "time_step": 1,
                    "yaw_rate": 0.0,
                    "slip_angle": 0.0,
                    "orientation": -np.pi / 2.,
                    "position": np.array([0.0, 0.0]),
                    "velocity": 0.01
                }),
                [1],
                False,
                np.array([2.01])
        )
    ]
)
@functional
@unit_test
def test_get_relative_future_goal_offsets(goal_region: GoalRegion, ego_state: State, sampling_points: List[int],
                                          static: bool, expected_offset):
    """unittest for LaneletNetworkObservation._get_relative_future_goal_offsets method"""
    configs = {"lanelet_configs": {}}

    planning_problem = PlanningProblem(planning_problem_id=0, initial_state=ego_state,
                                       goal_region=goal_region)

    lanelet = Lanelet(lanelet_id=0,
                      left_vertices=np.array([[0.0, 3.0], [10.0, 3.0]]),
                      center_vertices=np.array([[0.0, 0.0], [10.0, 0.0]]),
                      right_vertices=np.array([[0.0, -3.0], [10.0, -3.0]]))

    scenario = Scenario(dt=0.1, scenario_id=ScenarioID("DEU_TEST-1_1_T-1"))
    scenario.lanelet_network.add_lanelet(lanelet)

    route_planner = RoutePlanner(
        scenario,
        planning_problem,
        backend=RoutePlanner.Backend.NETWORKX_REVERSED,
        log_to_console=False,
    )

    route_candidates = route_planner.plan_routes()
    route = route_candidates.retrieve_best_route_by_orientation()

    navigator = Navigator(route)

    observation = LaneletNetworkObservation(configs)
    calculated_offset = observation._get_relative_future_goal_offsets(ego_state, sampling_points, static, navigator)[0]
    assert np.isclose(calculated_offset, expected_offset)


@pytest.mark.parametrize(
    ("strict_off_road_check", "check_circle_radius", "action", "expected_output"),
    [
        (True, 0, np.array([0.0, 0.0]), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        (True, 0, np.array([0.05, 0.0]), [0, 0, 0, 0, 0, 1, 1, 1, 1, 0]),
        (False, 0.1, np.array([0.05, 0.0]), [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
        (False, 0.5, np.array([0.05, 0.0]), [0, 0, 0, 0, 0, 0, 1, 0, 0, 0]),
        (False, 1.0, np.array([0.05, 0.0]), [0, 0, 0, 0, 0, 1, 1, 0, 0, 0]),
        (False, 1.5, np.array([0.05, 0.0]), [0, 0, 0, 0, 1, 1, 1, 1, 0, 0]),
    ],
)
@functional
@unit_test
def test_check_off_road(strict_off_road_check, check_circle_radius, action, expected_output):
    # manully reset the ego vehicle
    vehicle_action.reset(CustomState(**dummy_state), dt=0.1)

    # specify off-road check mode
    lanelet_observation.strict_off_road_check = strict_off_road_check
    lanelet_observation.non_strict_check_circle_radius = check_circle_radius

    steps = 20
    result = np.zeros(steps)

    for i in range(steps):
        vehicle_action.step(action)
        result[i] = lanelet_observation._check_is_off_road(vehicle_action.vehicle, road_edge)

    assert np.all(result[-10:] == expected_output)


@functional
@unit_test
def test_is_off_road_end():
    vehicle_action.reset(CustomState(position=np.array([387.5, 5.]),
                               orientation=0.,
                               velocity=0.,
                               time_step=0), dt=0.1)

    # specify off-road check mode
    lanelet_observation.strict_off_road_check = True

    # should not be off road if open end
    assert not lanelet_observation._check_is_off_road(vehicle_action.vehicle, road_edge)

    from commonroad_rl.tools.pickle_scenario.preprocessing import generate_reset_config
    reset_config = generate_reset_config(scenario, open_lane_ends=False)

    assert lanelet_observation._check_is_off_road(vehicle_action.vehicle, reset_config)


@module_test
@functional
def test_end_of_road_termination():
    env = gym.make("commonroad-v1",
                   meta_scenario_path=os.path.join(pickle_path, "meta_scenario"),
                   train_reset_config_path=os.path.join(pickle_path, "problem"))
    env.reset()
    # manually reset ego vehicle to the end of the road
    env.ego_action.reset(CustomState(position=np.array([387.5, 5.]),
                               orientation=0.,
                               velocity=1.,
                               time_step=0), dt=0.1)
    action = np.array([0., 0.])
    obs, reward, done, info = env.step(action)

    assert info["is_time_out"] == 1


@pytest.mark.parametrize(
    ("action", "steps", "expected_output"),
    [
        # (np.array([0., 0.]), 0, np.array([1.7479, 1.7479, 5.2813, 5.4769, ])),
        (np.array([0., 0.]), 5, np.array([1.9025, 1.5932, 5.4359, 5.3222, ])),
        (np.array([0., 0.]), 10, np.array([2.0571, 1.4385, 5.5906, 5.1675, ])),
        (np.array([0., 0.05]), 5, np.array([1.7571, 1.7386, 5.2905, 5.4676, ])),
        (np.array([0., 0.05]), 10, np.array([0.8946, 2.6011, 4.4280, 6.3301, ])),
    ],
)
@functional
@unit_test
def test_get_distance_to_marker_and_road_edge(action, steps, expected_output):
    # manully reset the ego vehicle
    vehicle_action.reset(CustomState(**dummy_state), dt=0.1)

    for i in range(steps):
        vehicle_action.step(action)

    # construct current lanelet of the ego vehicle
    ego_lanelet_ids = ObservationCollector.sorted_lanelets_by_state(scenario, vehicle_action.vehicle.state,
                                                                    lanelet_polygons, lanelet_polygons_sg)
    ego_lanelet_id = ego_lanelet_ids[0]
    ego_lanelet = scenario.lanelet_network.find_lanelet_by_id(ego_lanelet_id)
    lanelet_observation._get_distance_to_marker_and_road_edge(vehicle_action.vehicle.state, ego_lanelet, road_edge)

    result = np.hstack((lanelet_observation.observation_dict["left_marker_distance"],
                        lanelet_observation.observation_dict["right_marker_distance"],
                        lanelet_observation.observation_dict["left_road_edge_distance"],
                        lanelet_observation.observation_dict["right_road_edge_distance"]))

    assert np.allclose(result, expected_output, rtol=0.001)


@pytest.mark.parametrize(
    ("ego_vehicle_lanelet_id", "desired_left_edge", "desired_right_edge"),
    [
        (
                0,
                LineString([[0, 1], [3, 4]]),
                LineString([[112, 3], [23, 4]])
        ),
        (
                3,
                LineString([[0, 1], [3, 4]]),
                LineString([[112, 3], [23, 4]])
        ),
        (
                34,
                LineString([[0, 1], [3, 4], [2, 34]]),
                LineString([[112, 3], [23, 4], [2314, 34]])
        )
    ]
)
@functional
@unit_test
def test_get_road_edge(ego_vehicle_lanelet_id: int, desired_left_edge: LineString, desired_right_edge: LineString):
    road_edge = {
        "left_road_edge_lanelet_id_dict": np.zeros(ego_vehicle_lanelet_id + 1).astype(int),
        "right_road_edge_lanelet_id_dict": np.zeros(ego_vehicle_lanelet_id + 1).astype(int),
        "left_road_edge_dict": [desired_left_edge],
        "right_road_edge_dict": [desired_right_edge],
    }

    obs = LaneletNetworkObservation({"lanelet_configs": {}})
    left_road_edge, right_road_edge = obs._get_road_edge(road_edge, ego_vehicle_lanelet_id)
    assert all(np.isclose(left_road_edge.bounds, desired_left_edge.bounds))
    assert all(np.isclose(right_road_edge.bounds, desired_right_edge.bounds))


@pytest.mark.parametrize(
    ("position", "desired_lat_position", "arr"),
    [
        (
                np.array([136.1013, -52.64465]),
                0,
                np.array(
                    [
                        [136.1013, -52.64465],
                        [136.9463, -53.99465],
                        [138.9713, -55.94465],
                        [141.0563, -57.19465],
                        [144.0213, -58.59465],
                        [147.1013, -59.64465],
                        [152.3213, -60.39465],
                        [157.7213, -59.79465],
                        [163.6113, -58.04465],
                    ]
                ),
        ),
        (
                np.array([133.7763, -13.59465]),
                0.0008496,
                np.array(
                    [
                        [135.6863, 7.45535],
                        [135.0613, -1.19465],
                        [133.7763, -13.59465],
                        [132.1663, -27.94465],
                        [131.1563, -36.29465],
                    ]
                ),
        ),
        (
                np.array([135.2, 6]),
                -0.380153613,
                np.array(
                    [
                        [135.6863, 7.45535],
                        [135.0613, -1.19465],
                        [133.7763, -13.59465],
                        [132.1663, -27.94465],
                        [131.1563, -36.29465],
                    ]
                ),
        ),
        (
                np.array([137.1013, -54.64465]),
                -0.36073585,
                np.array(
                    [
                        [136.1013, -52.64465],
                        [136.9463, -53.99465],
                        [138.9713, -55.94465],
                        [141.0563, -57.19465],
                        [144.0213, -58.59465],
                        [147.1013, -59.64465],
                        [152.3213, -60.39465],
                        [157.7213, -59.79465],
                        [163.6113, -58.04465],
                    ]
                ),
        ),
    ]
)
@functional
@unit_test
def test_get_relative_offset(position: np.ndarray, desired_lat_position: float, arr: np.ndarray):
    curvi_cosy = Navigator.create_coordinate_system_from_polyline(arr)
    ego_vehicle_lat_position = LaneletNetworkObservation.get_relative_offset(curvi_cosy, position)
    assert np.isclose(ego_vehicle_lat_position, desired_lat_position), \
        f"differs: {np.array2string(np.array(ego_vehicle_lat_position), separator=',')} " \
        f"expected {np.array2string(np.array(desired_lat_position), separator=',')}: " \
        f"{np.isclose(np.array(ego_vehicle_lat_position), np.array(desired_lat_position), atol=1e-2)}"


@pytest.mark.parametrize(
    ("p", "l", "desired_distance"),
    [
        (
                Point(1, -1),
                LineString([[0, 0], [1, 0], [1, 1]]),
                1
        ),
        (
                Point(1, -1),
                LineString([[0, 0], [1, 0], [1, -1]]),
                0
        ),
        (
                Point(1, 2),
                LineString([[0, 0], [1, 0], [1, 1]]),
                1
        )
    ]
)
@functional
@unit_test
def test_get_distance_point_to_linestring(p: Point, l: LineString, desired_distance: float):
    observed_distance = LaneletNetworkObservation.get_distance_point_to_linestring(p, l)
    assert np.isclose(desired_distance, observed_distance)


@pytest.mark.parametrize(
    ("position", "left_marker", "right_marker", "left_edge", "right_edge", "desired_left_marker_distance",
     "desired_right_marker_distance", "desired_left_edge_distance", "desired_right_edge_distance"),
    [
        (
                np.array([0, 0]),
                LineString([[-1, 1.5], [0, 2], [1, 2]]),
                LineString([[-1, -1], [0, -1], [1, -2]]),
                LineString([[-1, 2], [0, 2.5], [1, 2.5]]),
                LineString([[-1, -1.5], [0, -1.5], [1, -1.5]]),
                1.7888543,
                1,
                2.2360679,
                1.5
        ),
        (
                np.array([1, 0.5]),
                LineString([[-1, 1.5], [0, 2], [1, 2]]),
                LineString([[-1, -1], [0, -1], [1, -2]]),
                LineString([[-1, 2], [0, 2.5], [1, 2.5]]),
                LineString([[-1, -1.5], [0, -1.5], [1, -1.5]]),
                1.5,
                1.80277563,
                2,
                2
        ),
        (
                np.array([1, 0.5]),
                LineString([[-0.5, 0.5], [0.5, 1.5]]),
                LineString([[-0.5, -0.5], [0.5, -0.5]]),
                LineString([[-0.5, 1.5], [0.5, 1.5]]),
                LineString([[-0.5, -1], [0.5, -1]]),
                1.06066017,
                1.11803398,
                1.11803398,
                1.58113883,
        ),
    ]
)
@functional
@unit_test
def test_get_distance_to_marker_and_road_edge(position: np.ndarray, left_marker: LineString, right_marker: LineString,
                                              left_edge: LineString, right_edge: LineString,
                                              desired_left_marker_distance: float, desired_right_marker_distance: float,
                                              desired_left_edge_distance: float, desired_right_edge_distance: float):
    ego_state = CustomState(**{"time_step": 0, "position": position, "orientation": 0, "velocity": 0})

    distance_left_marker, distance_right_marker, distance_left_road_edge, distance_right_road_edge = \
        LaneletNetworkObservation.get_distance_to_marker_and_road_edge(ego_state, left_marker, right_marker,
                                                                       left_edge, right_edge)

    assert np.isclose(desired_right_edge_distance, distance_right_road_edge)
    assert np.isclose(desired_left_edge_distance, distance_left_road_edge)
    assert np.isclose(desired_right_marker_distance, distance_right_marker)
    assert np.isclose(desired_left_marker_distance, distance_left_marker)


@pytest.mark.parametrize(
    ("position", "curve", "expected_output"),
    [
        (np.array([0, 1.0]),
         np.array([[0.70710678, 0.70710678],
                   [0.57357644, 0.81915204],
                   [0.42261826, 0.90630779],
                   [0.25881905, 0.96592583],
                   [0.08715574, 0.9961947],
                   [0, 1.0],
                   [-0.08715574, 0.9961947],
                   [-0.25881905, 0.96592583],
                   [-0.42261826, 0.90630779],
                   [-0.57357644, 0.81915204],
                   [-0.70710678, 0.70710678]]), 1.07),
        (np.array([0., 2.]), np.array([[1.41421356, 1.41421356],
                                       [1.14715287, 1.63830409],
                                       [0.84523652, 1.81261557],
                                       [0.51763809, 1.93185165],
                                       [0.17431149, 1.9923894],
                                       [-0.17431149, 1.9923894],
                                       [-0.51763809, 1.93185165],
                                       [-0.84523652, 1.81261557],
                                       [-1.14715287, 1.63830409],
                                       [-1.41421356, 1.41421356]]), 0.5),
        (np.array([0., 0.]), np.array([[0, 0],
                                       [1, 1],
                                       [2, 2],
                                       [3, 3],
                                       [4, 4]]), 0.0)
    ]
)
@functional
@unit_test
def test_get_lane_curvature(position, curve, expected_output):
    from commonroad_dc.geometry.geometry import CurvilinearCoordinateSystem # import here cause otherwise it's overwritten by pycrccosy.CurvilinearCoordinateSystem
    ccosy = CurvilinearCoordinateSystem(curve)
    curvature = LaneletNetworkObservation.get_lane_curvature(position, ccosy)
    # TODO: reduce tolerance after ccosy is fixed (extrapolation and resampling introduced curvature deviation)
    assert np.isclose(curvature, expected_output, atol=1e-2)



""" Module for managing the vehicle in the CommonRoad Gym environment
"""
import copy

import numpy as np
import commonroad_dc.pycrcc as pycrcc
from typing import List, Tuple, Union
from abc import ABC, abstractmethod
from aenum import extend_enum
from scipy.optimize import Bounds
from commonroad.scenario.trajectory import State, CustomState
from commonroad_rl.gym_commonroad.utils.scenario import make_valid_orientation
from commonroad.common.solution import VehicleModel, VehicleType
from commonroad_dc.collision.trajectory_queries import trajectory_queries
from commonroad_dc.feasibility.vehicle_dynamics import VehicleDynamics, FrictionCircleException
from vehiclemodels.vehicle_parameters import VehicleParameters
from vehiclemodels.parameters_vehicle1 import parameters_vehicle1
from vehiclemodels.parameters_vehicle2 import parameters_vehicle2
from vehiclemodels.parameters_vehicle3 import parameters_vehicle3
from scipy.integrate import odeint

N_INTEGRATION_STEPS = 100

extend_enum(VehicleModel, 'YawRate', len(VehicleModel))
extend_enum(VehicleModel, 'QP', len(VehicleModel))
extend_enum(VehicleModel, 'PMNonlinear', len(VehicleModel))


# Using VehicleParameterMapping from feasibility checker causes bugs
def to_vehicle_parameter(vehicle_type: VehicleType):
    if vehicle_type == VehicleType.FORD_ESCORT:
        return parameters_vehicle1()
    elif vehicle_type == VehicleType.BMW_320i:
        return parameters_vehicle2()
    elif vehicle_type == VehicleType.VW_VANAGON:
        return parameters_vehicle3()
    else:
        raise TypeError(f"Vehicle type {vehicle_type} not supported!")


def assert_vehicle_model(vehicle_model: VehicleModel):
    if vehicle_model == VehicleModel.MB:
        raise NotImplementedError(f"Vehicle model {vehicle_model} is not implemented yet!")
    else:
        return vehicle_model


class Vehicle(ABC):
    """
    Description:
        Abstract base class of all vehicles
    """

    def __init__(self, params_dict: dict) -> None:
        """ Initialize empty object """
        vehicle_type = VehicleType(params_dict["vehicle_type"])
        vehicle_model = VehicleModel(params_dict["vehicle_model"])

        self.vehicle_type = vehicle_type
        self.vehicle_model = assert_vehicle_model(vehicle_model)
        self.parameters = to_vehicle_parameter(vehicle_type)
        self.name = None
        self.dt = None
        self._collision_object = None
        self.initial_state = None
        self.state_list = None

    @property
    def state(self) -> State:
        """
        Get the current state of the vehicle

        :return: The current state of the vehicle
        """
        return self.state_list[-1]

    @property
    def previous_state(self) -> State:
        """
        Get the previous state of the vehicle

        :return: The previous state of the vehicle
        """
        if len(self.state_list) > 1:
            return self.state_list[-2]
        else:
            return self.initial_state

    @state.setter
    def state(self, state: State):
        """ Set the current state of the vehicle is not supported """
        raise ValueError("To set the state of the vehicle directly is prohibited!")

    @property
    def collision_object(self) -> pycrcc.RectOBB:
        """
        Get the collision object of the vehicle

        :return: The collision object of the vehicle
        """
        return self._collision_object

    @collision_object.setter
    def collision_object(self, collision_object: pycrcc.RectOBB):
        """ Set the collision_object of the vehicle is not supported """
        raise ValueError("To set the collision_object of the vehicle directly is prohibited!")

    @property
    def current_time_step(self):
        return self.state.time_step

    @current_time_step.setter
    def current_time_step(self, current_time_step):
        raise ValueError("To set the current time step of the vehicle directly is prohibited!")

    def create_obb_collision_object(self, state: State):
        return pycrcc.RectOBB(self.parameters.l / 2,
                              self.parameters.w / 2,
                              state.orientation,
                              state.position[0],
                              state.position[1])

    def update_collision_object(self, create_convex_hull=True):
        """ Updates the collision_object of the vehicle """
        if create_convex_hull:
            self._collision_object = pycrcc.TimeVariantCollisionObject(self.previous_state.time_step)
            self._collision_object.append_obstacle(self.create_obb_collision_object(self.previous_state))
            self._collision_object.append_obstacle(self.create_obb_collision_object(self.state))
            self._collision_object, err = trajectory_queries.trajectory_preprocess_obb_sum(self._collision_object)
            if not err:
                return
            raise Exception("trajectory preprocessing error")
        else:
            self._collision_object = pycrcc.TimeVariantCollisionObject(self.state.time_step)
            self._collision_object.append_obstacle(self.create_obb_collision_object(self.state))

    @abstractmethod
    def set_current_state(self, new_state: State):
        """
        Update state list
        """
        raise NotImplementedError

    def reset(self, initial_state: State, dt: float) -> None:
        """
        Reset vehicle parameters.

        :param initial_state: The initial state of the vehicle
        :param dt: Simulation dt of the scenario
        :return: None
        """
        self.dt = dt
        if self.vehicle_model == VehicleModel.PM:
            orientation = initial_state.orientation if hasattr(initial_state, "orientation") else 0.0
            self.initial_state = CustomState(**{"position": initial_state.position,
                                          "orientation": make_valid_orientation(orientation),
                                          "time_step": initial_state.time_step,
                                          "velocity": initial_state.velocity * np.cos(orientation),
                                          "velocity_y": initial_state.velocity * np.sin(orientation),
                                          "acceleration": initial_state.acceleration * np.cos(orientation)
                                          if hasattr(initial_state, "acceleration") else 0.0,
                                          "acceleration_y": initial_state.acceleration * np.sin(orientation)
                                          if hasattr(initial_state, "acceleration") else 0.0})
        elif self.vehicle_model == VehicleModel.QP:
            self.initial_state = CustomState(**{"position": initial_state.position,
                                          "velocity": initial_state.velocity,
                                          "acceleration": initial_state.acceleration
                                          if hasattr(initial_state, "acceleration")
                                          else 0.0,
                                          "jerk": initial_state.jerk
                                          if hasattr(initial_state, "jerk") else 0.0,
                                          "orientation": make_valid_orientation(initial_state.orientation)
                                          if hasattr(initial_state, "orientation")
                                          else 0.0,
                                          "slip_angle": 0.0,
                                          "yaw_rate": 0.0,
                                          "time_step": initial_state.time_step,
                                          })
        else:
            self.initial_state = CustomState(**{"position": initial_state.position,
                                          "steering_angle": initial_state.steering_angle
                                          if hasattr(initial_state, "steering_angle")
                                          else 0.0,
                                          "orientation": make_valid_orientation(initial_state.orientation)
                                          if hasattr(initial_state, "orientation")
                                          else 0.0,
                                          "yaw_rate": initial_state.yaw_rate
                                          if hasattr(initial_state, "yaw_rate")
                                          else 0.0,
                                          "time_step": initial_state.time_step,
                                          "velocity": initial_state.velocity,
                                          "acceleration": initial_state.acceleration
                                          if hasattr(initial_state, "acceleration")
                                          else 0.0})
        self.state_list: List[State] = [self.initial_state]
        self.update_collision_object(create_convex_hull=self._continuous_collision_checking)

    def rescale_action(self, normalized_action: np.ndarray) -> np.ndarray:
        """
        Rescales the normalized action from [-1,1] to the required range

        :param normalized_action: action from the CommonroadEnv.
        :return: rescaled action
        """
        pass


class ContinuousVehicle(Vehicle):
    """
    Description:
        Class for vehicle when trained in continuous action space
    """

    def __init__(self, params_dict: dict, continuous_collision_checking=True):
        """ Initialize empty object """
        super().__init__(params_dict)
        self.violate_friction = False
        self.jerk_bounds = np.array([-10000, 10000])
        self._continuous_collision_checking = continuous_collision_checking

        try:
            self.vehicle_dynamic = VehicleDynamics.from_model(self.vehicle_model, self.vehicle_type)
        except:
            if self.vehicle_model == VehicleModel.YawRate:
                # customize YawRate VehicleModel
                self.vehicle_dynamic = YawRateDynamics(self.vehicle_type)
                self.parameters = self.vehicle_dynamic.parameters
            elif self.vehicle_model == VehicleModel.QP:
                self.vehicle_dynamic = QPDynamics(self.vehicle_type)
            elif self.vehicle_model == VehicleModel.PMNonlinear:
                self.vehicle_dynamic = PointMassNonlinearDynamics(self.vehicle_type)
            else:
                raise ValueError(f"Unknown vehicle model: {self.vehicle_model}")

    def set_current_state(self, new_state: State):
        """
        Update state list

        :param new_state: new state
        :return: None
        """
        self.state_list.append(new_state)
        self.update_collision_object(create_convex_hull=self._continuous_collision_checking)

    def propagate_one_time_step(self, current_state: State, action: np.ndarray, action_base: str) -> State:
        """Generate the next state from a given state for the given action.

        :param current_state: current state of vehicle to propagate from
        :param action: control inputs of vehicle (real input)
        :param action_base: aspect on which the action should be based ("jerk", "acceleration")
        :return: propagated state
        """
        if action_base == "acceleration":
            u_input = action
        elif action_base == "jerk":
            u_input = self._jerk_to_acc(action)
        else:
            raise ValueError(f"Unknown action base: {action_base}")

        if self.vehicle_model == VehicleModel.PM:
            # using vehicle_dynamics.state_to_array(current_state) causes error since state has orientation and velocity
            x_current = np.array(
                [
                    current_state.position[0],
                    current_state.position[1],
                    current_state.velocity,
                    current_state.velocity_y,
                ]
            )

            # if maximum absolute acceleration is exceeded, rescale the acceleration
            absolute_acc = u_input[0] ** 2 + u_input[1] ** 2
            if absolute_acc > self.parameters.longitudinal.a_max ** 2:
                rescale_factor = (self.parameters.longitudinal.a_max - 1e-6) / np.sqrt(absolute_acc)
                # rescale the acceleration to satisfy friction circle constraint
                u_input[0] *= rescale_factor
                u_input[1] *= rescale_factor
        else:
            x_current, _ = self.vehicle_dynamic.state_to_array(current_state)

        try:
            x_current_old = copy.deepcopy(x_current)
            x_current = self.vehicle_dynamic.forward_simulation(x_current, u_input, self.dt, throw=True)
            self.violate_friction = False
        except FrictionCircleException:
            self.violate_friction = True
            for _ in range(N_INTEGRATION_STEPS):
                # simulate state transition - t parameter is set to vehicle.dt but irrelevant for the current vehicle models
                # TODO：x_dot of KS model considers the action constraints, which YR and PM model have not included yet
                x_dot = np.array(self.vehicle_dynamic.dynamics(self.dt, x_current, u_input))
                # update state
                x_current = x_current + x_dot * (self.dt / N_INTEGRATION_STEPS)

        # feed in required slots
        if self.vehicle_model == VehicleModel.PM:
            kwarg = {
                "position": np.array([x_current[0], x_current[1]]),
                "velocity": x_current[2],
                "velocity_y": x_current[3],
                "acceleration": u_input[0],
                "acceleration_y": u_input[1],
                "orientation": make_valid_orientation(np.arctan2(x_current[3], x_current[2])),
                "time_step": current_state.time_step + 1,
            }
        elif self.vehicle_model == VehicleModel.KS:
            state = self.vehicle_dynamic.array_to_state(x_current, time_step=current_state.time_step + 1)
            state.orientation = make_valid_orientation(state.orientation)
            state.acceleration = u_input[1]
            state.yaw_rate = (x_current[4] - x_current_old[4]) / self.dt

            return state
        elif self.vehicle_model == VehicleModel.YawRate:
            kwarg = {
                "position": np.array([x_current[0], x_current[1]]),
                "steering_angle": x_current[2],
                "velocity": x_current[3],
                "orientation": make_valid_orientation(x_current[4]),
                "acceleration": u_input[1],
                "yaw_rate": u_input[0],
                "time_step": current_state.time_step + 1,
            }
        elif self.vehicle_model == VehicleModel.QP:
            state = self.vehicle_dynamic.array_to_state(x_current, time_step=current_state.time_step + 1)
            # TODO velocity_z and roll_rate are used to jounce and kappa_dot_dot
            state.velocity_z = u_input[0]
            state.roll_rate = u_input[1]

            return state
        elif self.vehicle_model == VehicleModel.PMNonlinear:
            kwarg = {
                "position": np.array([x_current[0], x_current[1]]),
                "velocity": x_current[2],
                "acceleration": u_input[0],
                "orientation": make_valid_orientation(x_current[3]),
                "yaw_rate": u_input[1],
                "time_step": current_state.time_step + 1,
            }

        return CustomState(**kwarg)

    def get_new_state(self, action: np.ndarray, action_base: str) -> State:
        """Generate the next state from current state for the given action.

        :params action: rescaled action
        :params action_base: aspect on which the action should be based ("jerk", "acceleration")
        :return: next state of vehicle"""

        current_state = self.state

        return self.propagate_one_time_step(current_state, action, action_base)

    def _jerk_to_acc(self, action: np.ndarray) -> np.ndarray:
        """
        computes the acceleration based input on jerk based actions
        :param action: action based on jerk
        :return: input based on acceleration
        """
        if self.vehicle_model == VehicleModel.PM:
            # action[jerk_x, jerk_y]
            action = np.array([np.clip(action[0], self.jerk_bounds[0], self.jerk_bounds[1]),
                               np.clip(action[1], self.jerk_bounds[0], self.jerk_bounds[1])])
            u_input = np.array([self.state.acceleration + action[0] * self.dt,
                                self.state.acceleration_y + action[1] * self.dt])

        elif self.vehicle_model == VehicleModel.KS:
            # action[steering angel speed, jerk]
            action = np.array([action[0], np.clip(action[1], self.jerk_bounds[0], self.jerk_bounds[1])])
            u_input = np.array([action[0], self.state.acceleration + action[1] * self.dt])

        elif self.vehicle_model == VehicleModel.YawRate:
            # action[yaw, jerk]
            action = np.array([action[0], np.clip(action[1], self.jerk_bounds[0], self.jerk_bounds[1])])
            u_input = np.array([action[0], self.state.acceleration + action[1] * self.dt])
        else:
            raise ValueError(f"Unknown vehicle model: {self.vehicle_model}")

        return u_input


class YawParameters():
    def __init__(self):
        # constraints regarding yaw
        self.v_min = []  # minimum yaw velocity [rad/s]
        self.v_max = []  # maximum yaw velocity [rad/s]


def extend_vehicle_params(p: VehicleParameters) -> VehicleParameters:
    p.yaw = YawParameters()
    p.yaw.v_min = -2.  # minimum yaw velocity [rad/s]
    p.yaw.v_max = 2.  # maximum yaw velocity [rad/s]
    return p


class YawRateDynamics(VehicleDynamics):
    """
    Description:
        Class for the calculation of vehicle dynamics of YawRate vehicle model
    """

    def __init__(self, vehicle_type: VehicleType):
        super(YawRateDynamics, self).__init__(VehicleModel.YawRate, vehicle_type)
        self.parameters = extend_vehicle_params(self.parameters)
        self.wheelbase = self.parameters.a + self.parameters.b

        self.velocity = None

    def dynamics(self, t, x, u) -> List[float]:
        """
        Yaw Rate model dynamics function.

        :param x: state values, [position x, position y, steering angle, longitudinal velocity, orientation(yaw angle)]
        :param u: input values, [yaw rate, longitudinal acceleration]

        :return: system dynamics
        """
        velocity_x = x[3] * np.cos(x[4])
        velocity_y = x[3] * np.sin(x[4])
        self.velocity = x[3]

        # steering angle velocity depends on longitudinal velocity and yaw rate (as well as vehicle parameters)
        steering_ang_velocity = -u[0] * self.wheelbase / (x[3] ** 2 + u[0] * self.wheelbase ** 2)

        return [velocity_x, velocity_y, steering_ang_velocity, u[1], u[0]]

    @property
    def input_bounds(self) -> Bounds:
        """
        Overrides the bounds method of Vehicle Model in order to return bounds for the Yaw Rate Model inputs.

        Bounds are
            - -max longitudinal acc <= acceleration <= max longitudinal acc
            - mini yaw velocity <= yaw_rate <= max yaw velocity

        :return: Bounds
        """
        return Bounds([self.parameters.yaw.v_min - 1e-4, -self.parameters.longitudinal.a_max],
                      [self.parameters.yaw.v_max + 1e-4, self.parameters.longitudinal.a_max])

    def _state_to_array(self, state: State, steering_angle_default=0.0) -> Tuple[np.array, int]:
        """ Implementation of the VehicleDynamics abstract method. """
        values = [
            state.position[0],
            state.position[1],
            getattr(state, 'steering_angle', steering_angle_default),  # not defined in initial state
            state.velocity,
            state.orientation,
        ]
        return np.array(values), state.time_step

    def _array_to_state(self, x: np.array, time_step: int) -> State:
        """ Implementation of the VehicleDynamics abstract method. """
        values = {
            'position': np.array([x[0], x[1]]),
            'steering_angle': x[2],
            'velocity': x[3],
            'orientation': x[4],
        }
        state = CustomState(**values, time_step=time_step)
        return state

    def _input_to_array(self, input: State) -> Tuple[np.array, int]:
        """
        Actual conversion of input to array happens here. Vehicles can override this method to implement their own converter.
        """
        values = [
            input.yaw_rate,
            input.acceleration,
        ]
        return np.array(values), input.time_step

    def _array_to_input(self, u: np.array, time_step: int) -> State:
        """
        Actual conversion of input array to input happens here. Vehicles can override this method to implement their
        own converter.
        """
        values = {
            'yaw_rate': u[0],
            'acceleration': u[1],
        }
        return CustomState(**values, time_step=time_step)


class QPDynamics(VehicleDynamics):
    """
    Description:
        Class for the calculation of vehicle dynamics of YawRate vehicle model
    """

    def __init__(self, vehicle_type: VehicleType):
        super(QPDynamics, self).__init__(VehicleModel.QP, vehicle_type)
        self.theta_ref = 0.

    @property
    def input_bounds(self) -> Bounds:
        """
        Overrides the bounds method of Vehicle Model in order to return bounds for the Yaw Rate Model inputs.

        Bounds are
            - max jerk_dot <= jerk_dot <= max jerk_dot
            - min kappa_dot_dot <= kappa_dot_dot <= max kappa_dot_dot

        :return: Bounds
        """
        return Bounds(
            np.array([-1000., -20.]),
            np.array([1000., 20.])
        )

    @staticmethod
    def jerk_dot_constraints(jerk_dot, acceleration, p):
        if (jerk_dot < 0. and acceleration <= -p.a_max) or \
                (jerk_dot > 0. and acceleration >= p.a_max):
            jerk_dot = 0.
        # TODO: integrate jerk dot constrain in vehicle parameters
        return jerk_dot

    @staticmethod
    def kappa_dot_dot_constraints(kappa_dot_dot, kappa_dot, kappa_dot_min, kappa_dot_max):
        if (kappa_dot < kappa_dot_min and kappa_dot_dot < 0.) or (kappa_dot > kappa_dot_max and kappa_dot_dot > 0.):
            kappa_dot_dot = 0.
        # TODO: integrate jerk dot constrain in vehicle parameters
        return kappa_dot_dot

    @staticmethod
    def vehicle_dynamics_linear(x_init, u_init, kappa_dot_max, p, theta_ref):
        x = np.array([x_init[:4], x_init[4:]])
        u = []
        u.append(QPDynamics.jerk_dot_constraints(u_init[0], x[0, 2], p.longitudinal))
        u.append(QPDynamics.kappa_dot_dot_constraints(u_init[1], x[1, 3], -kappa_dot_max, kappa_dot_max))

        x_long = x[0, :]
        x_lat = x[1, :]

        # longitudinal dynamics
        f_long = [
            x_long[1],
            x_long[2],
            x_long[3],
            u[0]
        ]

        # lateral dynamics
        v = x_long[1]  # TODO: QP planner uses the propagated velocity

        f_lat = [
            v * x_lat[1] - v * theta_ref,
            v * x_lat[2],
            x_lat[3],
            u[1]
        ]

        # f = [f_long, f_lat]
        f = f_long + f_lat

        return f

    def discrete_dynamics(self, x, u_init, dt, kappa_dot_max):
        x = np.array([x[:4], x[4:]])
        u = []
        u.append(QPDynamics.jerk_dot_constraints(u_init[0], x[0, 2], self.parameters.longitudinal))
        u.append(QPDynamics.kappa_dot_dot_constraints(u_init[1], x[1, 3], -kappa_dot_max, kappa_dot_max))

        x_long = x[0, :]
        x_lat = x[1, :]

        A_long = np.array(
            [[1, dt, (dt ** 2.) / 2., (dt ** 3.) / 6.],
             [0, 1., dt, (dt ** 2.) / 2.],
             [0, 0, 1., dt],
             [0, 0, 0, 1]]
        )
        B_long = np.array([[(dt ** 4.) / 24.], [(dt ** 3.) / 6.], [(dt ** 2.) / 2.], [dt]])

        v = x_long[1]
        A_lat = np.array(
            [
                [1, dt * v, (dt ** 2) * 0.5 * (v ** 2), (dt ** 3) / 6 * (v ** 2)],
                [0, 1, dt * v, (dt ** 2) * 0.5 * v],
                [0, 0, 1, dt],
                [0, 0, 0, 1]
            ])

        B_lat = np.array([[(dt ** 4) / 24 * (v ** 2)],
                          [(dt ** 3) / 6 * v],
                          [(dt ** 2) * 0.5],
                          [dt]])

        x_long = np.dot(A_long, x_long) + np.squeeze(B_long * u[0])
        x_lat = np.dot(A_lat, x_lat) + np.squeeze(B_lat * u[1])

        return np.concatenate([x_long, x_lat])

    def dynamics(self, t, x, u_init, dt) -> List[float]:
        # wb = self.parameters.a + self.parameters.b
        # kappa_max = np.tan(self.parameters.steering.max) / wb
        kappa_max = 0.2
        kappa_dot_max = 2 * kappa_max / dt
        self.theta_ref = 0.  # TODO: integrate reference path
        return QPDynamics.vehicle_dynamics_linear(x, u_init, kappa_dot_max, self.parameters, self.theta_ref)

        # # discrete integration
        # return self.discrete_dynamics(x, u_init, dt, kappa_dot_max)

    def forward_simulation(self, x: np.array, u_init: np.array, dt: float, throw: bool = True) -> np.array:
        """
        Simulates the next state using the given state and input values as numpy arrays.

        :param x: state values.
        :param u: input values
        :param dt: scenario delta time.
        :param throw: if set to false, will return None as next state instead of throwing exception (default=True)
        :return: simulated next state values, raises VehicleDynamicsException if invalid input.
        """

        x0, x1 = odeint(self.dynamics, x, [0.0, dt], args=(u_init, dt, ), tfirst=True)
        # x1 = self.dynamics(None, x, u_init, dt)

        return x1

    def _state_to_array(self, state: State, steering_angle_default=0.0) -> Tuple[np.array, int]:
        """ Implementation of the VehicleDynamics abstract method. """
        # (s, v, a, j)  (d, θ, κ,˙κ)
        theta_ref = 0.
        # TODO: convert Cartesian position to CCOSY position
        values = np.array(
            [state.position[0] - self.parameters.b * np.cos(state.orientation),
             state.velocity * np.cos(state.orientation - self.theta_ref), state.acceleration, state.jerk] +
            [state.position[1] - self.parameters.b * np.sin(state.orientation), state.orientation, state.slip_angle,
             state.yaw_rate]
        )  # TODO slip_angle and yaw_rate are used to store kappa and kappa_dot

        return values, state.time_step

    def _array_to_state(self, x_init: np.array, time_step: int) -> State:
        """ Implementation of the VehicleDynamics abstract method. """
        # TODO: convert CCOSY position to Cartesian position
        # (s, v, a, j)  (d, θ, κ,˙κ)
        x = np.array([x_init[:4], x_init[4:]])
        values = {
            'position': np.array([x[0, 0] + self.parameters.b * np.cos(x[1, 1]),
                                  x[1, 0] + self.parameters.b * np.sin(x[1, 1])]),
            'velocity': x[0, 1] / np.cos(x[1, 1] - self.theta_ref),
            'acceleration': x[0, 2],
            'jerk': x[0, 3],
            'orientation': make_valid_orientation(x[1, 1]),
            'slip_angle': x[1, 2],
            'yaw_rate': x[1, 3],
        }
        state = CustomState(**values, time_step=time_step)
        return state


class PointMassNonlinearDynamics(VehicleDynamics):

    def __init__(self, vehicle_type: VehicleType):
        super(PointMassNonlinearDynamics, self).__init__(VehicleModel.PM, vehicle_type)

    def dynamics(self, t, x, u) -> List[float]:
        """
        Point Mass model dynamics function. Overrides the dynamics function of VehicleDynamics for PointMass model.

        :param t:
        :param x: state values, [position x, position y, velocity, orientation]
        :param u: input values, [acceleration, yaw rate]

        :return:
        """
        return [
            x[2] * np.cos(x[3]),
            x[2] * np.sin(x[3]),
            u[0],
            u[1]
        ]

    @property
    def input_bounds(self) -> Bounds:
        """
        Overrides the bounds method of Vehicle Model in order to return bounds for the Point Mass inputs.

        Bounds are
            - -max longitudinal acc <= acceleration <= max longitudinal acc
            - -max longitudinal acc <= acceleration_y <= max longitudinal acc

        :return: Bounds
        """
        return Bounds([-self.parameters.longitudinal.a_max, -self.parameters.longitudinal.a_max],
                      [self.parameters.longitudinal.a_max, self.parameters.longitudinal.a_max])

    def violates_friction_circle(self, x: Union[State, np.array], u: Union[State, np.array],
                                 throw: bool = False) -> bool:
        """
        Overrides the friction circle constraint method of Vehicle Model in order calculate
        friction circle constraint for the Point Mass model.

        :param x: current state
        :param u: the input which was used to simulate the next state
        :param throw: if set to false, will return bool instead of throwing exception (default=False)
        :return: True if the constraint was violated
        """
        u_vals = self.input_to_array(u)[0] if isinstance(u, State) else u
        x_vals = self.state_to_array(x)[0] if isinstance(x, State) else x

        a_long = u_vals[0]
        a_lat = u_vals[1] * x_vals[2]

        vals_power = a_long ** 2 + a_lat ** 2
        violates = vals_power > self.parameters.longitudinal.a_max ** 2

        if throw and violates:
            msg = f'Input violates friction circle constraint!\n' \
                  f'Init state: {x}\n\n Input:{u}'
            raise FrictionCircleException(msg)

        return violates

    def _state_to_array(self, state: State, steering_angle_default=0.0) -> Tuple[np.array, int]:
        """ Implementation of the VehicleDynamics abstract method. """
        values = [
            state.position[0],
            state.position[1],
            state.velocity,
            state.orientation
        ]
        time_step = state.time_step
        return np.array(values), time_step

    def _array_to_state(self, x: np.array, time_step: int) -> State:
        """ Implementation of the VehicleDynamics abstract method. """
        values = {
            'position': np.array([x[0], x[1]]),
            'velocity': x[2],
            'orientation': x[3]
        }
        return CustomState(**values, time_step=time_step)

    def _input_to_array(self, input: State) -> Tuple[np.array, int]:
        """ Overrides VehicleDynamics method. """
        values = [
            input.acceleration,
            input.yaw_rate,
        ]
        time_step = input.time_step
        return np.array(values), time_step

    def _array_to_input(self, u: np.array, time_step: int) -> State:
        """ Overrides VehicleDynamics method. """
        values = {
            'acceleration': u[0],
            'yaw_rate': u[1],
        }
        return CustomState(**values, time_step=time_step)

    def occupancy(self, x):
        """Compute the space occupied by the car, used only for the safety layer using reachable sets"""

        l = self.parameters.l
        w = self.parameters.w

        phi = x[3]
        R = np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])

        V = np.array([[l / 2, l / 2, -l / 2, -l / 2], [-w / 2, w / 2, w / 2, -w / 2]])

        V = np.dot(R, V) + np.reshape(x[0:2], (2, 1))

        return V


if __name__ == "__main__":
    continuous_vehicle = ContinuousVehicle()



"""
Module for playing trained model using Stable baselines
"""
import argparse
import csv
import time
import logging
import os
import yaml
import glob
import multiprocessing

os.environ["KMP_WARNINGS"] = "off"
os.environ["KMP_AFFINITY"] = "none"
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
os.environ["OMP_NUM_THREADS"] = "1"
logging.getLogger("tensorflow").disabled = True
from stable_baselines.common import set_global_seeds

from commonroad_rl.evaluate_model import create_environments
from commonroad_rl.train_model import LoggingMode
from commonroad_rl.utils_run.utils import load_model_and_vecnormalize
from commonroad_rl.gym_commonroad.observation.goal_observation import GoalObservation
from traci.exceptions import FatalTraCIError, TraCIException

import copy
import os
import pickle
from typing import Tuple, Dict
import numpy as np

from sumocr.sumo_config.default import DefaultConfig
from sumocr.interface.ego_vehicle import EgoVehicle
from sumocr.interface.sumo_simulation import SumoSimulation
from sumocr.maps.sumo_scenario import ScenarioWrapper

from commonroad.scenario.scenario import Scenario
from commonroad.planning.planning_problem import PlanningProblemSet, PlanningProblem
from commonroad.common.file_reader import CommonRoadFileReader
from commonroad_rl.gym_commonroad.commonroad_env import CommonroadEnv

try:
    from cbf_iss.cbf.cbf_wrapper import CBFWrapper
except ImportError:
    CBFWrapper = None
try:
    from mpi4py import MPI
except ImportError:
    print("ImportFailure MPI")
    MPI = None

LOGGER = logging.getLogger(__name__)


def get_parser():
    parser = argparse.ArgumentParser(description="Evaluates PPO2 trained model with SUMO interactive scenarios",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--env_id", type=str, default="commonroad-v1", help="environment ID")
    parser.add_argument("--algo", type=str, default="ppo2")
    parser.add_argument("--num_processes", "-n", type=int, default=1)
    parser.add_argument("--test_path", "-path", type=str, default="", help="Path to SUMO scenario folders")
    parser.add_argument("--evaluation_path", type=str, default="evaluation_sumo")
    parser.add_argument("--model_path", "-model", type=str, default="")
    parser.add_argument("--viz_path", "-viz", type=str, default="")
    parser.add_argument("--hyperparam_filename", "-hyperparam_f", type=str, default="model_hyperparameters.yml")
    parser.add_argument("--config_filename", "-config_f", type=str, default="environment_configurations.yml")
    parser.add_argument("--logging_mode", default=LoggingMode.INFO, type=LoggingMode, choices=list(LoggingMode))
    parser.add_argument("--render", "-r", action="store_true", help="Whether to render the env during simulation")

    return parser


def simulate_scenario(conf: DefaultConfig,
                      scenario_wrapper: ScenarioWrapper,
                      planning_problem_set: PlanningProblemSet = None,
                      model=None,
                      env=None, render=False) -> Tuple[Scenario, Dict[int, EgoVehicle], Dict]:
    """
    Simulates an interactive scenario with specified mode

    :param conf: config of the simulation
    :param scenario_wrapper: scenario wrapper used by the Simulator
    :param planning_problem_set: planning problem set of the scenario
    :return: simulated scenario and dictionary with items {planning_problem_id: EgoVehicle}
    """

    # initialize simulation
    sumo_sim = SumoSimulation()
    # TODO: fix in scenario converter
    conf.dt = scenario_wrapper.initial_scenario.dt
    sumo_sim.initialize(conf, scenario_wrapper, planning_problem_set=planning_problem_set)

    # simulation with plugged in planner
    if isinstance(env.venv.envs[0].env.env, CommonroadEnv):
        cr_env = env.venv.envs[0].env.env
    elif isinstance(env.venv.envs[0].env, CBFWrapper):
        cr_env = env.venv.envs[0].env.env.env

    assert isinstance(cr_env, CommonroadEnv)

    def run_simulation():
        step = 0
        done = False
        infos = dict()
        while not done:
            # retrieve the CommonRoad scenario at the current time step, e.g. as an input for a prediction module
            current_scenario = sumo_sim.commonroad_scenario_at_time_step(sumo_sim.current_time_step)
            # TODO: use commonroad-io Scenario.assign_obstacles_to_lanelets after next release
            current_scenario.assign_obstacles_to_lanelets(time_steps=[0])
            for idx, ego_vehicle in enumerate(sumo_sim.ego_vehicles.values()):
                # retrieve the current state of the ego vehicle
                planning_problem = list(planning_problem_set.planning_problem_dict.values())[idx]
                planning_initital_state = copy.deepcopy(ego_vehicle.current_state)
                # TODO: check if slip_angle matters
                planning_initital_state.time_step = 0
                planning_initital_state.slip_angle = 0.

                observations = env.reset(
                    scenario=current_scenario,
                    planning_problem=PlanningProblem(
                        planning_problem_id=planning_problem.planning_problem_id,
                        initial_state=planning_initital_state,
                        goal_region=planning_problem.goal
                    )
                )
                # fix is_time_out here since the current_step inside CommonRoadEnv is incorrect
                is_goal_reached = cr_env.observation_dict["is_goal_reached"][0]
                is_time_out = step >= cr_env.observation_collector.episode_length and not is_goal_reached
                cr_env.observation_dict["is_time_out"] = np.array([is_time_out])
                done, reason, termination_info = cr_env.termination.is_terminated(
                    cr_env.observation_dict,
                    cr_env.ego_action
                )

                infos.update(termination_info)

                # update infos (filter out invalid collision)
                if isinstance(env.venv.envs[0].env, CBFWrapper) and \
                        (termination_info["is_collision"] == 1 or termination_info["is_off_road"] == 1):
                    # rerun simulation to filter out invalid collision and offroad
                    # (should reproduce the same result since reproducable)
                    simulated_scenario = sumo_sim.commonroad_scenarios_all_time_steps()
                    obs = env.reset(
                        scenario=simulated_scenario,
                        planning_problem=planning_problem
                    )
                    done = False
                    while not done:
                        action = model.predict(obs, deterministic=True)
                        obs, reward, done, info = env.step(action)
                    infos = info[0]

                if render:
                    env.render(mode=step + 1)

                if done:
                    return infos # terminated by collision

                observation_keys = []
                for key, value in cr_env.observation_dict.items():
                    observation_keys.extend([key]*value.size)

                obs_distance_goal_long_advance_idx = observation_keys.index("distance_goal_long_advance")
                obs_distance_goal_lat_advance_idx = observation_keys.index("distance_goal_lat_advance")
                obs_distance_goal_time_idx = observation_keys.index("distance_goal_time")
                obs_remaining_steps_idx = observation_keys.index("remaining_steps")

                if step > 0:
                    # fix distance_goal_long_advance
                    distance_goal_long_advance = abs(last_distance_goal_long) - abs(
                        cr_env.observation_dict["distance_goal_long"])
                    distance_goal_lat_advance = abs(last_distance_goal_lat) - abs(
                        cr_env.observation_dict["distance_goal_lat"])

                    remaining_steps = cr_env.observation_dict["remaining_steps"] - step
                    distance_goal_time = np.array([GoalObservation._get_goal_time_distance(step, planning_problem.goal)])

                    observations[0][obs_distance_goal_long_advance_idx] = distance_goal_long_advance
                    observations[0][obs_distance_goal_lat_advance_idx] = distance_goal_lat_advance
                    observations[0][obs_remaining_steps_idx] = remaining_steps
                    observations[0][obs_distance_goal_time_idx] = distance_goal_time

                    # normalize
                    tmp_normalized_obs = env.normalize_obs(observations)
                    for idx in [obs_distance_goal_long_advance_idx, obs_distance_goal_lat_advance_idx, obs_remaining_steps_idx, obs_distance_goal_time_idx]:
                        observations[0][idx] = tmp_normalized_obs[0][idx]

                assert cr_env.scenario is current_scenario
                action = model.predict(observations, deterministic=True)

                last_distance_goal_long = cr_env.observation_dict["distance_goal_long"]
                last_distance_goal_lat = cr_env.observation_dict["distance_goal_lat"]
                _, _, done, info = env.step(action)
                infos = info[0]

                if done:
                    return infos # terminated by other reasons (offroad, timeout, goal reached etc)

                # get ego state
                next_state = cr_env.ego_action.vehicle.state

                # update the ego vehicle with new trajectory with only 1 state for the current step
                next_state.time_step = 1
                trajectory_ego = [next_state]
                ego_vehicle.set_planned_trajectory(trajectory_ego)

            if done:
                break
            sumo_sim.simulate_step()
            step += 1

    info = run_simulation()

    # retrieve the simulated scenario in CR format
    simulated_scenario = sumo_sim.commonroad_scenarios_all_time_steps()

    # stop the simulation
    sumo_sim.stop()

    ego_vehicles = {list(planning_problem_set.planning_problem_dict.keys())[0]:
                        ego_v for _, ego_v in sumo_sim.ego_vehicles.items()}

    return simulated_scenario, ego_vehicles, info


def simulate_with_planner(interactive_scenario_path: str,
                          create_ego_obstacle: bool = False,
                          model=None,
                          env=None,
                          render=False) \
        -> Tuple[Scenario, PlanningProblemSet, Dict[int, EgoVehicle]]:
    """
    Simulates an interactive scenario with a plugged in motion planner

    :param interactive_scenario_path: path to the interactive scenario folder
    :param output_folder_path: path to the output folder
    :param create_video: indicates whether to create a mp4 of the simulated scenario
    :param use_sumo_manager: indicates whether to use the SUMO Manager
    :param create_ego_obstacle: indicates whether to create obstacles from the planned trajectories as the ego vehicles
    :return: Tuple of the simulated scenario, planning problem set, and list of ego vehicles
    """
    conf = load_sumo_configuration(interactive_scenario_path)
    scenario_file = os.path.join(interactive_scenario_path, f"{conf.scenario_name}.cr.xml")
    scenario, planning_problem_set = CommonRoadFileReader(scenario_file).open()

    scenario_wrapper = ScenarioWrapper()
    scenario_wrapper.sumo_cfg_file = os.path.join(interactive_scenario_path, f"{conf.scenario_name}.sumo.cfg")
    scenario_wrapper.initial_scenario = scenario

    scenario_with_planner, ego_vehicles, info = simulate_scenario(
        conf,
        scenario_wrapper,
        planning_problem_set=planning_problem_set,
        model=model,
        env=env,
        render=render
    )

    scenario_with_planner.scenario_id = scenario.scenario_id

    if create_ego_obstacle:
        for pp_id, planning_problem in planning_problem_set.planning_problem_dict.items():
            obstacle_ego = ego_vehicles[pp_id].get_dynamic_obstacle()
            scenario_with_planner.add_objects(obstacle_ego)

    return scenario_with_planner, planning_problem_set, ego_vehicles, info


def load_sumo_configuration(interactive_scenario_path: str) -> DefaultConfig:
    with open(os.path.join(interactive_scenario_path, "simulation_config.p"), "rb") as input_file:
        conf = pickle.load(input_file)

    return conf


def simulate_batch(i, sumo_files, viz_path, evaluation_path, args):
    env, normalize = create_environments(
        args.env_id,
        args.model_path,
        viz_path,
        args.hyperparam_filename,
        args.config_filename,
        args.logging_mode.value,
        play=False,
        test_env=True,
        cache_navigators=True, # SUMO scenarios have to cache navigators since reset is called for each step
        action_configs={"continuous_collision_checking": False} # SUMO scenario don't support continuous collision check
    )
    model, env = load_model_and_vecnormalize(args.model_path, args.algo, normalize, env)

    fd_result = open(os.path.join(evaluation_path, f"{i}_results.csv"), "w")
    csv_writer = csv.writer(fd_result)

    num_valid_collisions, num_collisions, num_valid_off_road, num_offroad, \
    num_goal_reaching, num_timeout, total_scenarios = 0, 0, 0, 0, 0, 0, 0

    for i, sumo_fn in enumerate(sumo_files):
        # if os.getlogin() == "wangx":
        #     scenario_id = os.path.splitext(os.path.basename(sumo_fn))[0].replace("T-1", "I-1")
        #     sumo_fn = os.path.join(
        #         "/home/wangx/data/highD-dataset-v1.0/10m_goal_no_downsample_lane_change/sumo",
        #         scenario_id
        #     )
        try:
            scenario_with_planner, _, ego_vehicles, info = simulate_with_planner(
                sumo_fn, model=model, env=env, render=args.render
            )
        except (FatalTraCIError, TraCIException, AttributeError) as e:
            # copy erronous scenario
            os.makedirs(os.path.join(args.model_path, "error"), exist_ok=True)
            os.system(f"cp -r {sumo_fn} {os.path.join(args.model_path, 'error/.')}")
            LOGGER.info(f"scenario {sumo_fn}: {e}")
            continue

        # log collision rate, off-road rate, and goal-reaching rate
        total_scenarios += 1
        num_valid_collisions += info.get("valid_collision", info["is_collision"])
        num_collisions += info["is_collision"]
        num_timeout += info["is_time_out"]
        num_valid_off_road += info.get("valid_off_road", info["is_off_road"])
        num_offroad += info["is_off_road"]
        num_goal_reaching += info["is_goal_reached"]

        termination_reason = "other"
        if info.get("is_time_out", 0) == 1:
            termination_reason = "time_out"
        elif info.get("is_off_road", 0) == 1:
            if "valid_off_road" in info and info["valid_off_road"] == 1:
                termination_reason = "valid_off_road"
            else:
                termination_reason = "off_road"
        elif info.get("is_collision", 0) == 1:
            if "valid_collision" in info and info["valid_collision"] == 1:
                termination_reason = "valid_collision"
            else:
                termination_reason = "collision"
        elif info.get("is_goal_reached", 0) == 1:
            termination_reason = "goal_reached"

        if "scenario_name" not in info or "current_episode_time_step" not in info:
            print(f"info={info}")
            continue
        csv_writer.writerow((info["scenario_name"], info["current_episode_time_step"], termination_reason))

    fd_result.close()

    return num_valid_collisions, num_collisions, num_valid_off_road, num_offroad, \
           num_goal_reaching, num_timeout, total_scenarios


def main():
    t1 = time.time()
    args = get_parser().parse_args()

    LOGGER.setLevel(args.logging_mode.value)
    handler = logging.StreamHandler()
    handler.setLevel(args.logging_mode.value)
    LOGGER.addHandler(handler)

    LOGGER.info("Start")

    # TODO: take all sumo folders that have a pickle file in problem_test
    sumo_paths = sorted(glob.glob(os.path.join(args.test_path, "*.pickle")))

    # if os.getlogin() == "xiao":
    sumo_paths = [x[0] for x in os.walk(args.test_path)][1:]

    # create evaluation folder in model_path
    evaluation_path = os.path.join(args.model_path, args.evaluation_path)
    os.makedirs(evaluation_path, exist_ok=True)
    if args.viz_path == "":
        args.viz_path = os.path.join(evaluation_path, "img")

    LOGGER.info(f"Number of scenarios: {len(sumo_paths)}")

    set_global_seeds(1)

    if args.num_processes == 1:
        t1 = time.time()
        total_num_valid_collisions, total_num_collisions, total_num_valid_off_road, total_num_offroad, \
        total_num_goal_reaching, total_num_timeout, total_scenarios = \
            simulate_batch(0, sumo_paths, args.viz_path, evaluation_path, args)
        LOGGER.info(f"Elapsed time for simulating {len(sumo_paths)} scenarios: {time.time() - t1}s")
    else:
        num_files_per_process = len(sumo_paths) // args.num_processes
        with multiprocessing.Pool(processes=args.num_processes) as pool:
            results = pool.starmap(
                simulate_batch,
                [
                    (
                        i,
                        sumo_paths[i * num_files_per_process: (i + 1) * num_files_per_process],
                        args.viz_path,
                        evaluation_path,
                        args
                    )
                    for i in range(args.num_processes)
                ])

        total_num_valid_collisions, total_num_collisions, total_num_valid_off_road, total_num_offroad, \
        total_num_goal_reaching, total_num_timeout, total_scenarios = 0, 0, 0, 0, 0, 0, 0
        for res in results:
            num_valid_collisions, num_collisions, num_valid_off_road, num_offroad, \
            num_goal_reaching, num_timeout, num_scenarios = res
            total_num_valid_collisions += num_valid_collisions
            total_num_valid_off_road += num_valid_off_road
            total_num_collisions += num_collisions
            total_num_offroad += num_offroad
            total_num_goal_reaching += num_goal_reaching
            total_num_timeout += num_timeout
            total_scenarios += num_scenarios

    # save evaluation results
    with open(os.path.join(evaluation_path, "results.csv"), 'w') as fd_result:
        fd_result.write("benchmark_id, time_steps, termination_reason\n")
        for i in range(args.num_processes):
            path = os.path.join(evaluation_path, f"{i}_results.csv")
            with open(path, 'r') as f:
                fd_result.write(f.read())
            os.remove(path)

    with open(os.path.join(evaluation_path, "overview.yml"), "w") as f:
        yaml.dump({
            "total_scenarios": total_scenarios,
            "num_valid_collisions": total_num_valid_collisions,
            "num_collisions": total_num_collisions,
            "num_timeout": total_num_timeout,
            "num_valid_off_road": total_num_valid_off_road,
            "num_off_road": total_num_offroad,
            "num_goal_reached": total_num_goal_reaching,
            "percentage_goal_reached": 100.0 * total_num_goal_reaching / total_scenarios,
            "percentage_offroad": 100.0 * total_num_offroad / total_scenarios,
            "percentage_collisions": 100.0 * total_num_collisions / total_scenarios,
            "percentage_valid_off_road": 100.0 * total_num_valid_off_road / total_scenarios,
            "percentage_valid_collisions": 100.0 * total_num_valid_collisions / total_scenarios,
            "percentage_timeout": 100.0 * total_num_timeout / total_scenarios
        }, f)

    print(f"Elapsed time for {len(sumo_paths)} scenarios: {time.time()-t1}s")


if __name__ == "__main__":
    main()
